@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2022 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 /* Conversion functions: C to Java. These all throw fitz exceptions. */
 
 static inline jobject to_ColorSpace(fz_context *ctx, JNIEnv *env, fz_colorspace *cs)
@@ -178,6 +200,20 @@ static inline jfloatArray to_floatArray(fz_context *ctx, JNIEnv *env, const floa
 
 /* Conversion functions: C to Java. None of these throw fitz exceptions. */
 
+static inline jobject to_Buffer_safe(fz_context *ctx, JNIEnv *env, fz_buffer *buf)
+{
+	jobject jbuf;
+
+	if (!ctx || !buf) return NULL;
+
+	fz_keep_buffer(ctx, buf);
+	jbuf = (*env)->NewObject(env, cls_Buffer, mid_Buffer_init, jlong_cast(buf));
+	if (!jbuf)
+		fz_drop_buffer(ctx, buf);
+
+	return jbuf;
+}
+
 static inline jint to_ColorParams_safe(fz_context *ctx, JNIEnv *env, fz_color_params cp)
 {
 	if (!ctx) return 0;
@@ -296,6 +332,18 @@ static inline jobject to_Outline_safe(fz_context *ctx, JNIEnv *env, fz_document
 	return jarr;
 }
 
+static inline jobject to_OutlineIterator_safe(fz_context *ctx, JNIEnv *env, fz_outline_iterator *iterator)
+{
+	jobject jiterator = NULL;
+
+	if (!ctx || !iterator) return NULL;
+
+	jiterator = (*env)->NewObject(env, cls_OutlineIterator, mid_OutlineIterator_init, (jlong)iterator);
+	if (!jiterator || (*env)->ExceptionCheck(env)) return NULL;
+
+	return jiterator;
+}
+
 static inline jobject to_PDFAnnotation_safe(fz_context *ctx, JNIEnv *env, pdf_annot *annot)
 {
 	jobject jannot;
@@ -310,17 +358,31 @@ static inline jobject to_PDFAnnotation_safe(fz_context *ctx, JNIEnv *env, pdf_an
 	return jannot;
 }
 
-static inline jobject to_PDFObject_safe(fz_context *ctx, JNIEnv *env, jobject pdf, pdf_obj *obj)
+static inline jobject to_PDFDocument_safe(fz_context *ctx, JNIEnv *env, pdf_document *pdf)
+{
+	jobject jpdf;
+
+	if (!ctx || !pdf) return NULL;
+
+	pdf_keep_document(ctx, pdf);
+	jpdf = (*env)->NewObject(env, cls_PDFDocument, mid_PDFDocument_init, jlong_cast(pdf));
+	if (!jpdf)
+		pdf_drop_document(ctx, pdf);
+
+	return jpdf;
+}
+
+static inline jobject to_PDFObject_safe(fz_context *ctx, JNIEnv *env, pdf_obj *obj)
 {
 	jobject jobj;
 
-	if (!ctx || !pdf) return NULL;
+	if (!ctx) return NULL;
 
 	if (obj == NULL)
 		return (*env)->GetStaticObjectField(env, cls_PDFObject, fid_PDFObject_Null);
 
 	pdf_keep_obj(ctx, obj);
-	jobj = (*env)->NewObject(env, cls_PDFObject, mid_PDFObject_init, jlong_cast(obj), pdf);
+	jobj = (*env)->NewObject(env, cls_PDFObject, mid_PDFObject_init, jlong_cast(obj));
 	if (!jobj)
 		pdf_drop_obj(ctx, obj);
 
@@ -369,6 +431,65 @@ static inline jobjectArray to_QuadArray_safe(fz_context *ctx, JNIEnv *env, const
 	return arr;
 }
 
+static int count_next_hits(const int *marks, int a, int end)
+{
+	int b = a + 1;
+	while (b < end && !marks[b])
+		++b;
+	return b - a;
+}
+
+/* Array of Array of Quad */
+static inline jobjectArray to_SearchHits_safe(fz_context *ctx, JNIEnv *env, const int *marks, const fz_quad *quads, jint n)
+{
+	jobjectArray toparr, arr;
+	int i, k, a, m;
+
+	if (!ctx || !marks || !quads)
+		return NULL;
+
+	/* Count total number of search hits */
+	i = a = 0;
+	while (a < n)
+	{
+		a += count_next_hits(marks, a, n);
+		++i;
+	}
+
+	toparr = (*env)->NewObjectArray(env, i, cls_ArrayOfQuad, NULL);
+	if (!toparr || (*env)->ExceptionCheck(env))
+		return NULL;
+
+	i = a = 0;
+	while (a < n)
+	{
+		m = count_next_hits(marks, a, n);
+
+		arr = (*env)->NewObjectArray(env, m, cls_Quad, NULL);
+		if (!arr || (*env)->ExceptionCheck(env))
+			return NULL;
+		(*env)->SetObjectArrayElement(env, toparr, i++, arr);
+		if ((*env)->ExceptionCheck(env))
+			return NULL;
+
+		for (k = 0; k < m; ++k) {
+			jobject jquad = to_Quad_safe(ctx, env, quads[a+k]);
+			if (!jquad || (*env)->ExceptionCheck(env))
+				return NULL;
+			(*env)->SetObjectArrayElement(env, arr, k, jquad);
+			if ((*env)->ExceptionCheck(env))
+				return NULL;
+			(*env)->DeleteLocalRef(env, jquad);
+		}
+
+		(*env)->DeleteLocalRef(env, arr);
+
+		a += m;
+	}
+
+	return toparr;
+}
+
 static inline jobject to_Rect_safe(fz_context *ctx, JNIEnv *env, fz_rect rect)
 {
 	if (!ctx) return NULL;
@@ -402,7 +523,7 @@ static inline jobjectArray to_StringArray_safe(fz_context *ctx, JNIEnv *env, con
 	return arr;
 }
 
-static inline jobject to_PDFWidget_safe(fz_context *ctx, JNIEnv *env, pdf_widget *widget)
+static inline jobject to_PDFWidget_safe(fz_context *ctx, JNIEnv *env, pdf_annot *widget)
 {
 	jobject jwidget;
 	int nopts;
@@ -422,7 +543,7 @@ static inline jobject to_PDFWidget_safe(fz_context *ctx, JNIEnv *env, pdf_widget
 	fz_try(ctx)
 	{
 		int fieldType = pdf_widget_type(ctx, widget);
-		int fieldFlags = pdf_field_flags(ctx, widget->obj);
+		int fieldFlags = pdf_field_flags(ctx, pdf_annot_obj(ctx, widget));
 		(*env)->SetIntField(env, jwidget, fid_PDFWidget_fieldType, fieldType);
 		(*env)->SetIntField(env, jwidget, fid_PDFWidget_fieldFlags, fieldFlags);
 		if (fieldType == PDF_WIDGET_TYPE_TEXT)
@@ -477,19 +598,6 @@ static inline jobject to_Document_safe_own(fz_context *ctx, JNIEnv *env, fz_docu
 	return obj;
 }
 
-static inline jobject to_Device_safe_own(fz_context *ctx, JNIEnv *env, fz_device *device)
-{
-	jobject jdev;
-
-	if (!ctx || !device) return NULL;
-
-	jdev = (*env)->NewObject(env, cls_DisplayList, mid_Device_init, jlong_cast(device));
-	if (!jdev)
-		fz_drop_device(ctx, device);
-
-	return jdev;
-}
-
 static inline jobject to_DisplayList_safe_own(fz_context *ctx, JNIEnv *env, fz_display_list *list)
 {
 	jobject jlist;
@@ -503,6 +611,19 @@ static inline jobject to_DisplayList_safe_own(fz_context *ctx, JNIEnv *env, fz_d
 	return jlist;
 }
 
+static inline jobject to_NativeDevice_safe_own(fz_context *ctx, JNIEnv *env, fz_device *device)
+{
+	jobject jdev;
+
+	if (!ctx || !device) return NULL;
+
+	jdev = (*env)->NewObject(env, cls_NativeDevice, mid_NativeDevice_init, jlong_cast(device));
+	if (!jdev)
+		fz_drop_device(ctx, device);
+
+	return jdev;
+}
+
 static inline jobject to_Page_safe_own(fz_context *ctx, JNIEnv *env, fz_page *page)
 {
 	jobject jobj;
@@ -521,6 +642,35 @@ static inline jobject to_Page_safe_own(fz_context *ctx, JNIEnv *env, fz_page *pa
 	return jobj;
 }
 
+static inline jobject to_Link_safe_own(fz_context *ctx, JNIEnv *env, fz_link *link)
+{
+	jobject jobj;
+	jobject jbounds = NULL;
+	jobject juri = NULL;
+
+	if (!ctx || !link) return NULL;
+
+	jbounds = to_Rect_safe(ctx, env, link->rect);
+	if (!jbounds || (*env)->ExceptionCheck(env))
+	{
+		fz_drop_link(ctx, link);
+		return NULL;
+	}
+
+	juri = (*env)->NewStringUTF(env, link->uri);
+	if (!juri || (*env)->ExceptionCheck(env))
+	{
+		fz_drop_link(ctx, link);
+		return NULL;
+	}
+
+	jobj = (*env)->NewObject(env, cls_Link, mid_Link_init, jbounds, juri);
+	if (!jobj)
+		fz_drop_link(ctx, link);
+
+	return jobj;
+}
+
 static inline jobject to_PDFAnnotation_safe_own(fz_context *ctx, JNIEnv *env, pdf_annot *annot)
 {
 	jobject jannot;
@@ -534,6 +684,19 @@ static inline jobject to_PDFAnnotation_safe_own(fz_context *ctx, JNIEnv *env, pd
 	return jannot;
 }
 
+static inline jobject to_PDFWidget_safe_own(fz_context *ctx, JNIEnv *env, pdf_annot *widget)
+{
+	jobject jwidget;
+
+	if (!ctx || !widget) return NULL;
+
+	jwidget = (*env)->NewObject(env, cls_PDFWidget, mid_PDFWidget_init, jlong_cast(widget));
+	if (!jwidget)
+		pdf_drop_annot(ctx, widget);
+
+	return jwidget;
+}
+
 static inline jobject to_PDFGraftMap_safe_own(fz_context *ctx, JNIEnv *env, jobject pdf, pdf_graft_map *map)
 {
 	jobject jmap;
@@ -547,13 +710,13 @@ static inline jobject to_PDFGraftMap_safe_own(fz_context *ctx, JNIEnv *env, jobj
 	return jmap;
 }
 
-static inline jobject to_PDFObject_safe_own(fz_context *ctx, JNIEnv *env, jobject pdf, pdf_obj *obj)
+static inline jobject to_PDFObject_safe_own(fz_context *ctx, JNIEnv *env, pdf_obj *obj)
 {
 	jobject jobj;
 
-	if (!ctx || !obj || !pdf) return NULL;
+	if (!ctx || !obj) return NULL;
 
-	jobj = (*env)->NewObject(env, cls_PDFObject, mid_PDFObject_init, jlong_cast(obj), pdf);
+	jobj = (*env)->NewObject(env, cls_PDFObject, mid_PDFObject_init, jlong_cast(obj));
 	if (!jobj)
 		pdf_drop_obj(ctx, obj);
 
@@ -669,6 +832,15 @@ static inline fz_image *from_Image(JNIEnv *env, jobject jobj)
 	return image;
 }
 
+static inline fz_outline_iterator *from_OutlineIterator(JNIEnv *env, jobject jobj)
+{
+	fz_outline_iterator *iterator;
+	if (!jobj) return NULL;
+	iterator = CAST(fz_outline_iterator *, (*env)->GetLongField(env, jobj, fid_OutlineIterator_pointer));
+	if (!iterator) jni_throw_null(env, "cannot use already destroyed OutlineIterator");
+	return iterator;
+}
+
 static inline fz_page *from_Page(JNIEnv *env, jobject jobj)
 {
 	fz_page *page;
@@ -864,6 +1036,24 @@ static inline fz_quad from_Quad(JNIEnv *env, jobject jquad)
 	return quad;
 }
 
+static fz_link_dest from_LinkDestination(JNIEnv *env, jobject jdest)
+{
+	fz_link_dest dest;
+
+	if (!jdest)
+		return fz_make_link_dest_none();
+
+	dest.loc.chapter = (*env)->GetIntField(env, jdest, fid_LinkDestination_chapter);
+	dest.loc.page = (*env)->GetIntField(env, jdest, fid_LinkDestination_page);
+	dest.type = (*env)->GetIntField(env, jdest, fid_LinkDestination_type);
+	dest.x = (*env)->GetFloatField(env, jdest, fid_LinkDestination_x);
+	dest.y = (*env)->GetFloatField(env, jdest, fid_LinkDestination_y);
+	dest.w = (*env)->GetFloatField(env, jdest, fid_LinkDestination_width);
+	dest.h = (*env)->GetFloatField(env, jdest, fid_LinkDestination_height);
+	dest.zoom = (*env)->GetFloatField(env, jdest, fid_LinkDestination_zoom);
+
+	return dest;
+}
 
 /* Conversion functions: Java to C. None of these throw java exceptions. */
 
@@ -969,10 +1159,10 @@ static inline pdf_obj *from_PDFObject_safe(JNIEnv *env, jobject jobj)
 	return CAST(pdf_obj *, (*env)->GetLongField(env, jobj, fid_PDFObject_pointer));
 }
 
-static inline pdf_widget *from_PDFWidget_safe(JNIEnv *env, jobject jobj)
+static inline pdf_annot *from_PDFWidget_safe(JNIEnv *env, jobject jobj)
 {
 	if (!jobj) return NULL;
-	return CAST(pdf_widget *, (*env)->GetLongField(env, jobj, fid_PDFWidget_pointer));
+	return CAST(pdf_annot *, (*env)->GetLongField(env, jobj, fid_PDFWidget_pointer));
 }
 
 static inline pdf_pkcs7_signer *from_PKCS7Signer_safe(JNIEnv *env, jobject jobj)
