@@ -1,11 +1,38 @@
+// Copyright (C) 2004-2022 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "emscripten.h"
 #include "mupdf/fitz.h"
+#include <string.h>
+#include <math.h>
 
 static fz_context *ctx;
 
 void wasm_rethrow(fz_context *ctx)
 {
-	EM_ASM({ throw new Error(UTF8ToString($0)); }, fz_caught_message(ctx));
+	if (fz_caught(ctx) == FZ_ERROR_TRYLATER)
+		EM_ASM({ throw "trylater"; });
+	else
+		EM_ASM({ throw new Error(UTF8ToString($0)); }, fz_caught_message(ctx));
 }
 
 EMSCRIPTEN_KEEPALIVE
@@ -13,14 +40,12 @@ void initContext(void)
 {
 	ctx = fz_new_context(NULL, NULL, 100<<20);
 	if (!ctx)
-	{
 		EM_ASM({ throw new Error("Cannot create MuPDF context!"); });
-	}
 	fz_register_document_handlers(ctx);
 }
 
 EMSCRIPTEN_KEEPALIVE
-fz_document *openDocumentFromBuffer(char *magic, unsigned char *data, size_t len)
+fz_document *openDocumentFromBuffer(unsigned char *data, int size, char *magic)
 {
 	fz_document *document = NULL;
 	fz_buffer *buf = NULL;
@@ -33,7 +58,7 @@ fz_document *openDocumentFromBuffer(char *magic, unsigned char *data, size_t len
 
 	fz_try(ctx)
 	{
-		buf = fz_new_buffer_from_data(ctx, data, len);
+		buf = fz_new_buffer_from_data(ctx, data, size);
 		stm = fz_open_buffer(ctx, buf);
 		document = fz_open_document_with_stream(ctx, magic, stm);
 	}
@@ -50,6 +75,17 @@ fz_document *openDocumentFromBuffer(char *magic, unsigned char *data, size_t len
 	return document;
 }
 
+EMSCRIPTEN_KEEPALIVE
+fz_document *openDocumentFromStream(fz_stream *stm, char *magic)
+{
+	fz_document *document = NULL;
+	fz_try(ctx)
+		document = fz_open_document_with_stream(ctx, magic, stm);
+	fz_catch(ctx)
+		wasm_rethrow(ctx);
+	return document;
+}
+
 EMSCRIPTEN_KEEPALIVE
 void freeDocument(fz_document *doc)
 {
@@ -292,7 +328,7 @@ char *search(fz_document *doc, int number, float dpi, const char *needle)
 	{
 		loadPage(doc, number);
 
-		n = fz_search_page(ctx, lastPage, needle, hits, nelem(hits));
+		n = fz_search_page(ctx, lastPage, needle, NULL, hits, nelem(hits));
 
 		buf = fz_new_buffer(ctx, 0);
 
@@ -368,9 +404,9 @@ char *outlineTitle(fz_outline *node)
 }
 
 EMSCRIPTEN_KEEPALIVE
-int outlinePage(fz_outline *node)
+int outlinePage(fz_document *doc, fz_outline *node)
 {
-	return node->page + 1;
+	return fz_page_number_from_location(ctx, doc, node->page);
 }
 
 EMSCRIPTEN_KEEPALIVE
@@ -384,3 +420,131 @@ fz_outline *outlineNext(fz_outline *node)
 {
 	return node->next;
 }
+
+/* PROGRESSIVE FETCH STREAM */
+
+struct fetch_state
+{
+	int block_shift;
+	int block_size;
+	int content_length; // Content-Length in bytes
+	int map_length; // Content-Length in blocks
+	uint8_t *content; // Array buffer with bytes
+	uint8_t *map; // Map of which blocks have been requested and loaded.
+};
+
+static void fetch_close(fz_context *ctx, void *state_)
+{
+	struct fetch_state *state = state_;
+	fz_free(ctx, state->content);
+	fz_free(ctx, state->map);
+	state->content = NULL;
+	state->map = NULL;
+	// TODO: wait for all outstanding requests to complete, then free state
+	// fz_free(ctx, state);
+	EM_ASM({
+		fetchClose($0);
+	}, state);
+}
+
+static void fetch_seek(fz_context *ctx, fz_stream *stm, int64_t offset, int whence)
+{
+	struct fetch_state *state = stm->state;
+	stm->wp = stm->rp = state->content;
+	if (whence == SEEK_END)
+		stm->pos = state->content_length + offset;
+	else if (whence == SEEK_CUR)
+		stm->pos += offset;
+	else
+		stm->pos = offset;
+	if (stm->pos < 0)
+		stm->pos = 0;
+	if (stm->pos > state->content_length)
+		stm->pos = state->content_length;
+}
+
+static int fetch_next(fz_context *ctx, fz_stream *stm, size_t len)
+{
+	struct fetch_state *state = stm->state;
+
+	int block = stm->pos >> state->block_shift;
+	int start = block << state->block_shift;
+	int end = start + state->block_size;
+	if (end > state->content_length)
+		end = state->content_length;
+
+	if (state->map[block] == 0) {
+		state->map[block] = 1;
+		EM_ASM({
+			fetchRead($0, $1);
+		}, state, block);
+		fz_throw(ctx, FZ_ERROR_TRYLATER, "waiting for data");
+	}
+
+	if (state->map[block] == 1) {
+		fz_throw(ctx, FZ_ERROR_TRYLATER, "waiting for data");
+	}
+
+	stm->rp = state->content + stm->pos;
+	stm->wp = state->content + end;
+	stm->pos = end;
+
+	if (stm->rp < stm->wp)
+		return *stm->rp++;
+	return -1;
+}
+
+EM_JS(void, js_open_fetch, (struct fetch_state *state, char *url, int content_length, int block_shift, int prefetch), {
+	fetchOpen(state, UTF8ToString(url), content_length, block_shift, prefetch);
+});
+
+EMSCRIPTEN_KEEPALIVE
+void onFetchData(struct fetch_state *state, int block, uint8_t *data, int size)
+{
+	if (state->content) {
+		memcpy(state->content + (block << state->block_shift), data, size);
+		state->map[block] = 2;
+	}
+}
+
+EMSCRIPTEN_KEEPALIVE
+fz_stream *openURL(char *url, int content_length, int block_size, int prefetch)
+{
+	fz_stream *stm = NULL;
+	struct fetch_state *state = NULL;
+	fz_var(stm);
+	fz_var(state);
+	fz_try (ctx)
+	{
+		int block_shift = (int)log2(block_size);
+
+		if (block_shift < 10 || block_shift > 24)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "invalid block shift: %d", block_shift);
+
+		state = fz_malloc(ctx, sizeof *state);
+		state->block_shift = block_shift;
+		state->block_size = 1 << block_shift;
+		state->content_length = content_length;
+		state->content = fz_malloc(ctx, state->content_length);
+		state->map_length = content_length / state->block_size + 1;
+		state->map = fz_malloc(ctx, state->map_length);
+		memset(state->map, 0, state->map_length);
+
+		stm = fz_new_stream(ctx, state, fetch_next, fetch_close);
+		// stm->progressive = 1;
+		stm->seek = fetch_seek;
+
+		js_open_fetch(state, url, content_length, block_shift, prefetch);
+	}
+	fz_catch(ctx)
+	{
+		if (state)
+		{
+			fz_free(ctx, state->content);
+			fz_free(ctx, state->map);
+			fz_free(ctx, state);
+		}
+		wasm_rethrow(ctx);
+	}
+	return stm;
+}
