@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 #include "draw-imp.h"
 
@@ -216,7 +238,7 @@ static void fz_postindex_edgebuffer(fz_context *ctx, fz_rasterizer *r)
 
 	if (eb->table_cap < total)
 	{
-		eb->table = fz_resize_array(ctx, eb->table, total, sizeof(int));
+		eb->table = fz_realloc_array(ctx, eb->table, total, int);
 		eb->table_cap = total;
 	}
 
@@ -236,7 +258,7 @@ static int fz_reset_edgebuffer(fz_context *ctx, fz_rasterizer *r)
 
 	if (eb->index_cap < height)
 	{
-		eb->index = fz_resize_array(ctx, eb->index, height, sizeof(int));
+		eb->index = fz_realloc_array(ctx, eb->index, height, int);
 		eb->index_cap = height;
 	}
 	memset(eb->index, 0, sizeof(int) * height);
