@@ -1,6 +1,30 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/ucdn.h"
 
+#include "glyphbox.h"
+
 #include <math.h>
 #include <float.h>
 #include <string.h>
@@ -32,12 +56,12 @@ void fz_drop_layout(fz_context *ctx, fz_layout_block *block)
 		fz_drop_pool(ctx, block->pool);
 }
 
-void fz_add_layout_line(fz_context *ctx, fz_layout_block *block, float x, float y, float h, const char *p)
+void fz_add_layout_line(fz_context *ctx, fz_layout_block *block, float x, float y, float font_size, const char *p)
 {
 	fz_layout_line *line = fz_pool_alloc(ctx, block->pool, sizeof (fz_layout_line));
 	line->x = x;
 	line->y = y;
-	line->h = h;
+	line->font_size = font_size;
 	line->p = p;
 	line->text = NULL;
 	line->next = NULL;
@@ -46,11 +70,11 @@ void fz_add_layout_line(fz_context *ctx, fz_layout_block *block, float x, float
 	block->text_tailp = &line->text;
 }
 
-void fz_add_layout_char(fz_context *ctx, fz_layout_block *block, float x, float w, const char *p)
+void fz_add_layout_char(fz_context *ctx, fz_layout_block *block, float x, float advance, const char *p)
 {
 	fz_layout_char *ch = fz_pool_alloc(ctx, block->pool, sizeof (fz_layout_char));
 	ch->x = x;
-	ch->w = w;
+	ch->advance = advance;
 	ch->p = p;
 	ch->next = NULL;
 	*block->text_tailp = ch;
@@ -85,6 +109,7 @@ const char *fz_stext_options_usage =
 	"\tpreserve-whitespace: do not convert all whitespace into space characters\n"
 	"\tpreserve-spans: do not merge spans on the same line\n"
 	"\tdehyphenate: attempt to join up hyphenated words\n"
+	"\tmediabox-clip=no: include characters outside mediabox\n"
 	"\n";
 
 fz_stext_page *
@@ -114,9 +139,17 @@ fz_drop_stext_page(fz_context *ctx, fz_stext_page *page)
 	if (page)
 	{
 		fz_stext_block *block;
+		fz_stext_line *line;
+		fz_stext_char *ch;
 		for (block = page->first_block; block; block = block->next)
+		{
 			if (block->type == FZ_STEXT_BLOCK_IMAGE)
 				fz_drop_image(ctx, block->u.i.image);
+			else
+				for (line = block->u.t.first_line; line; line = line->next)
+					for (ch = line->first_char; ch; ch = ch->next)
+						fz_drop_font(ctx, ch->font);
+		}
 		fz_drop_pool(ctx, page->pool);
 	}
 }
@@ -125,6 +158,7 @@ static fz_stext_block *
 add_block_to_page(fz_context *ctx, fz_stext_page *page)
 {
 	fz_stext_block *block = fz_pool_alloc(ctx, page->pool, sizeof *page->first_block);
+	block->bbox = fz_empty_rect; /* Fixes bug 703267. */
 	block->prev = page->last_block;
 	if (!page->first_block)
 		page->first_block = page->last_block = block;
@@ -192,7 +226,7 @@ add_char_to_line(fz_context *ctx, fz_stext_page *page, fz_stext_line *line, fz_m
 	ch->color = color;
 	ch->origin = *p;
 	ch->size = size;
-	ch->font = font; /* TODO: keep and drop */
+	ch->font = fz_keep_font(ctx, font);
 
 	if (line->wmode == 0)
 	{
@@ -203,9 +237,8 @@ add_char_to_line(fz_context *ctx, fz_stext_page *page, fz_stext_line *line, fz_m
 	}
 	else
 	{
-		fz_rect bbox = fz_font_bbox(ctx, font);
-		a.x = bbox.x1;
-		d.x = bbox.x0;
+		a.x = 1;
+		d.x = 0;
 		a.y = 0;
 		d.y = 0;
 	}
@@ -301,7 +334,6 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 	fz_point delta;
 	float spacing = 0;
 	float base_offset = 0;
-	int rtl = 0;
 
 	dev->curdir = direction_from_bidi_class(ucdn_get_bidi_class(c), dev->curdir);
 
@@ -425,10 +457,12 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 			else
 			{
 				new_line = 0;
+#if 0 /* TODO: handle RTL visual/logical ordering */
 				if (spacing > size * SPACE_DIST || spacing < 0)
 					rtl = 0; /* backward (or big jump to 'right' side) means logical order */
 				else
 					rtl = 1; /* visual order, we need to reverse in a post process pass */
+#endif
 			}
 		}
 
@@ -572,6 +606,10 @@ fz_stext_extract(fz_context *ctx, fz_stext_device *dev, fz_text_span *span, fz_m
 		tm.f = span->items[i].y;
 		trm = fz_concat(tm, ctm);
 
+		if (dev->flags & FZ_STEXT_MEDIABOX_CLIP)
+			if (fz_glyph_entirely_outside_box(ctx, &ctm, span, &span->items[i], &dev->page->mediabox))
+				continue;
+
 		/* Calculate bounding box and new pen position based on font metrics */
 		if (span->items[i].gid >= 0)
 			adv = fz_advance_glyph(ctx, font, span->items[i].gid, span->wmode);
@@ -716,7 +754,7 @@ fz_new_image_from_shade(fz_context *ctx, fz_shade *shade, fz_matrix *in_out_ctm,
 			fz_fill_pixmap_with_color(ctx, pix, shade->colorspace, shade->background, color_params);
 		else
 			fz_clear_pixmap(ctx, pix);
-		fz_paint_shade(ctx, shade, NULL, ctm, pix, color_params, bbox, NULL);
+		fz_paint_shade(ctx, shade, NULL, ctm, pix, color_params, bbox, NULL, NULL);
 		img = fz_new_image_from_pixmap(ctx, pix, NULL);
 	}
 	fz_always(ctx)
@@ -805,6 +843,14 @@ fz_parse_stext_options(fz_context *ctx, fz_stext_options *opts, const char *stri
 	if (fz_has_option(ctx, string, "preserve-spans", &val) && fz_option_eq(val, "yes"))
 		opts->flags |= FZ_STEXT_PRESERVE_SPANS;
 
+	opts->flags |= FZ_STEXT_MEDIABOX_CLIP;
+	if (fz_has_option(ctx, string, "mediabox-clip", &val) && fz_option_eq(val, "no"))
+		opts->flags ^= FZ_STEXT_MEDIABOX_CLIP;
+
+	opts->scale = 1;
+	if (fz_has_option(ctx, string, "resolution", &val))
+		opts->scale = fz_atof(val) / 96.0f; /* HTML base resolution is 96ppi */
+
 	return opts;
 }
 
