@@ -1,46 +1,61 @@
-#include "fitz-imp.h"
-#include "glyph-cache-imp.h"
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
+#include "mupdf/fitz.h"
 
 #include <string.h>
 #include <float.h>
+#include <math.h>
 
-typedef struct svg_device_s svg_device;
-
-typedef struct tile_s tile;
-typedef struct font_s font;
-typedef struct glyph_s glyph;
-typedef struct image_s image;
-
-struct tile_s
+typedef struct
 {
 	int pattern;
 	fz_matrix ctm;
 	fz_rect view;
 	fz_rect area;
 	fz_point step;
-};
+} tile;
 
-struct glyph_s
+typedef struct glyph
 {
 	float x_off;
 	float y_off;
-};
+} glyph;
 
-struct font_s
+typedef struct
 {
 	int id;
 	fz_font *font;
 	int max_sentlist;
 	glyph *sentlist;
-};
+} font;
 
-struct image_s
+typedef struct
 {
 	int id;
 	fz_image *image;
-};
+} image;
 
-struct svg_device_s
+typedef struct
 {
 	fz_device super;
 
@@ -53,8 +68,11 @@ struct svg_device_s
 	fz_buffer *defs_buffer;
 	int def_count;
 
+	int *save_id;
 	int id;
 
+	int blend_bitmask;
+
 	int num_tiles;
 	int max_tiles;
 	tile *tiles;
@@ -68,7 +86,10 @@ struct svg_device_s
 	image *images;
 
 	int layers;
-};
+
+	float page_width;
+	float page_height;
+} svg_device;
 
 /* SVG is awkward about letting us define things within symbol definitions
  * so we have to delay definitions until after the symbol definition ends. */
@@ -156,19 +177,19 @@ svg_dev_path(fz_context *ctx, svg_device *sdev, const fz_path *path)
 }
 
 static void
-svg_dev_ctm(fz_context *ctx, svg_device *sdev, const fz_matrix *ctm)
+svg_dev_ctm(fz_context *ctx, svg_device *sdev, fz_matrix ctm)
 {
 	fz_output *out = sdev->out;
 
-	if (ctm->a != 1.0f || ctm->b != 0 || ctm->c != 0 || ctm->d != 1.0f || ctm->e != 0 || ctm->f != 0)
+	if (ctm.a != 1.0f || ctm.b != 0 || ctm.c != 0 || ctm.d != 1.0f || ctm.e != 0 || ctm.f != 0)
 	{
 		fz_write_printf(ctx, out, " transform=\"matrix(%g,%g,%g,%g,%g,%g)\"",
-			ctm->a, ctm->b, ctm->c, ctm->d, ctm->e, ctm->f);
+			ctm.a, ctm.b, ctm.c, ctm.d, ctm.e, ctm.f);
 	}
 }
 
 static void
-svg_dev_stroke_state(fz_context *ctx, svg_device *sdev, const fz_stroke_state *stroke_state, const fz_matrix *ctm)
+svg_dev_stroke_state(fz_context *ctx, svg_device *sdev, const fz_stroke_state *stroke_state, fz_matrix ctm)
 {
 	fz_output *out = sdev->out;
 	float exp;
@@ -177,8 +198,6 @@ svg_dev_stroke_state(fz_context *ctx, svg_device *sdev, const fz_stroke_state *s
 	if (exp == 0)
 		exp = 1;
 	exp = stroke_state->linewidth/exp;
-	if (exp < 1)
-		exp = 1;
 
 	fz_write_printf(ctx, out, " stroke-width=\"%g\"", exp);
 	fz_write_printf(ctx, out, " stroke-linecap=\"%s\"",
@@ -202,14 +221,14 @@ svg_dev_stroke_state(fz_context *ctx, svg_device *sdev, const fz_stroke_state *s
 }
 
 static unsigned int
-svg_hex_color(fz_context *ctx, fz_colorspace *colorspace, const float *color, const fz_color_params *color_params)
+svg_hex_color(fz_context *ctx, fz_colorspace *colorspace, const float *color, fz_color_params color_params)
 {
 	float rgb[3];
 	int r, g, b;
 
 	if (colorspace != fz_device_rgb(ctx))
 	{
-		fz_convert_color(ctx, color_params, NULL, fz_device_rgb(ctx), rgb, colorspace, color);
+		fz_convert_color(ctx, colorspace, color, fz_device_rgb(ctx), rgb, NULL, color_params);
 		color = rgb;
 	}
 
@@ -221,7 +240,7 @@ svg_hex_color(fz_context *ctx, fz_colorspace *colorspace, const float *color, co
 }
 
 static void
-svg_dev_fill_color(fz_context *ctx, svg_device *sdev, fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+svg_dev_fill_color(fz_context *ctx, svg_device *sdev, fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	fz_output *out = sdev->out;
 	if (colorspace)
@@ -237,7 +256,7 @@ svg_dev_fill_color(fz_context *ctx, svg_device *sdev, fz_colorspace *colorspace,
 }
 
 static void
-svg_dev_stroke_color(fz_context *ctx, svg_device *sdev, fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+svg_dev_stroke_color(fz_context *ctx, svg_device *sdev, fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	fz_output *out = sdev->out;
 	if (colorspace)
@@ -269,13 +288,13 @@ find_first_char(fz_context *ctx, const fz_text_span *span, int i)
 }
 
 static int
-find_next_line_break(fz_context *ctx, const fz_text_span *span, const fz_matrix *inv_tm, int i)
+find_next_line_break(fz_context *ctx, const fz_text_span *span, fz_matrix inv_tm, int i)
 {
 	fz_point p, old_p;
 
 	old_p.x = span->items[i].x;
 	old_p.y = span->items[i].y;
-	fz_transform_point(&old_p, inv_tm);
+	old_p = fz_transform_point(old_p, inv_tm);
 
 	for (++i; i < span->len; ++i)
 	{
@@ -283,7 +302,7 @@ find_next_line_break(fz_context *ctx, const fz_text_span *span, const fz_matrix
 		{
 			p.x = span->items[i].x;
 			p.y = span->items[i].y;
-			fz_transform_point(&p, inv_tm);
+			p = fz_transform_point(p, inv_tm);
 			if (span->wmode == 0)
 			{
 				if (p.y != old_p.y)
@@ -313,7 +332,7 @@ svg_cluster_advance(fz_context *ctx, const fz_text_span *span, int i, int end)
 }
 
 static void
-svg_dev_text_span(fz_context *ctx, svg_device *sdev, const fz_matrix *ctm, const fz_text_span *span)
+svg_dev_text_span(fz_context *ctx, svg_device *sdev, fz_matrix ctm, const fz_text_span *span)
 {
 	fz_output *out = sdev->out;
 	char font_family[100];
@@ -332,15 +351,15 @@ svg_dev_text_span(fz_context *ctx, svg_device *sdev, const fz_matrix *ctm, const
 	}
 
 	tm = span->trm;
-	font_size = fz_matrix_expansion(&tm);
+	font_size = fz_matrix_expansion(tm);
 	final_tm.a = tm.a / font_size;
 	final_tm.b = tm.b / font_size;
 	final_tm.c = -tm.c / font_size;
 	final_tm.d = -tm.d / font_size;
 	final_tm.e = 0;
 	final_tm.f = 0;
-	fz_invert_matrix(&inv_tm, &final_tm);
-	fz_concat(&final_tm, &final_tm, ctm);
+	inv_tm = fz_invert_matrix(final_tm);
+	final_tm = fz_concat(final_tm, ctm);
 
 	tm.e = span->items[0].x;
 	tm.f = span->items[0].y;
@@ -362,11 +381,11 @@ svg_dev_text_span(fz_context *ctx, svg_device *sdev, const fz_matrix *ctm, const
 	start = find_first_char(ctx, span, 0);
 	while (start < span->len)
 	{
-		end = find_next_line_break(ctx, span, &inv_tm, start);
+		end = find_next_line_break(ctx, span, inv_tm, start);
 
 		p.x = span->items[start].x;
 		p.y = span->items[start].y;
-		fz_transform_point(&p, &inv_tm);
+		p = fz_transform_point(p, inv_tm);
 		if (span->items[start].gid >= 0)
 			cluster_advance = svg_cluster_advance(ctx, span, start, end);
 		if (span->wmode == 0)
@@ -384,7 +403,7 @@ svg_dev_text_span(fz_context *ctx, svg_device *sdev, const fz_matrix *ctm, const
 				{
 					p.x = it->x;
 					p.y = it->y;
-					fz_transform_point(&p, &inv_tm);
+					p = fz_transform_point(p, inv_tm);
 				}
 				else
 				{
@@ -419,7 +438,7 @@ svg_dev_text_span(fz_context *ctx, svg_device *sdev, const fz_matrix *ctm, const
 }
 
 static font *
-svg_dev_text_span_as_paths_defs(fz_context *ctx, fz_device *dev, fz_text_span *span, const fz_matrix *ctm)
+svg_dev_text_span_as_paths_defs(fz_context *ctx, fz_device *dev, fz_text_span *span, fz_matrix ctm)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out = sdev->out;
@@ -440,8 +459,8 @@ svg_dev_text_span_as_paths_defs(fz_context *ctx, fz_device *dev, fz_text_span *s
 			int newmax = sdev->max_fonts * 2;
 			if (newmax == 0)
 				newmax = 4;
-			sdev->fonts = fz_resize_array(ctx, sdev->fonts, newmax, sizeof(*sdev->fonts));
-			memset(&sdev->fonts[font_idx], 0, (newmax - font_idx) * sizeof(sdev->fonts[0]));
+			sdev->fonts = fz_realloc_array(ctx, sdev->fonts, newmax, font);
+			memset(&sdev->fonts[font_idx], 0, (newmax - font_idx) * sizeof(font));
 			sdev->max_fonts = newmax;
 		}
 		sdev->fonts[font_idx].id = sdev->id++;
@@ -460,7 +479,7 @@ svg_dev_text_span_as_paths_defs(fz_context *ctx, fz_device *dev, fz_text_span *s
 		if (gid >= fnt->max_sentlist)
 		{
 			int j;
-			fnt->sentlist = fz_resize_array(ctx, fnt->sentlist, gid+1, sizeof(fnt->sentlist[0]));
+			fnt->sentlist = fz_realloc_array(ctx, fnt->sentlist, gid+1, glyph);
 			for (j = fnt->max_sentlist; j <= gid; j++)
 			{
 				fnt->sentlist[j].x_off = FLT_MIN;
@@ -471,30 +490,36 @@ svg_dev_text_span_as_paths_defs(fz_context *ctx, fz_device *dev, fz_text_span *s
 		if (fnt->sentlist[gid].x_off == FLT_MIN)
 		{
 			/* Need to send this one */
-			fz_rect rect;
+			fz_rect rect = fz_empty_rect;
 			fz_path *path;
-			path = fz_outline_glyph(ctx, span->font, gid, &fz_identity);
-			if (path)
+			out = start_def(ctx, sdev);
+			fz_write_printf(ctx, out, "<symbol id=\"font_%x_%x\">\n", fnt->id, gid);
+			if (fz_font_ft_face(ctx, span->font))
 			{
-				fz_bound_path(ctx, path, NULL, &fz_identity, &rect);
-				shift.e = -rect.x0;
-				shift.f = -rect.y0;
-				fz_transform_path(ctx, path, &shift);
-				out = start_def(ctx, sdev);
-				fz_write_printf(ctx, out, "<symbol id=\"font_%x_%x\">\n", fnt->id, gid);
-				fz_write_printf(ctx, out, "<path");
-				svg_dev_path(ctx, sdev, path);
-				fz_write_printf(ctx, out, "/>\n");
-				fz_drop_path(ctx, path);
+				path = fz_outline_glyph(ctx, span->font, gid, fz_identity);
+				if (path)
+				{
+					rect = fz_bound_path(ctx, path, NULL, fz_identity);
+					shift.e = -rect.x0;
+					shift.f = -rect.y0;
+					fz_transform_path(ctx, path, shift);
+					fz_write_printf(ctx, out, "<path");
+					svg_dev_path(ctx, sdev, path);
+					fz_write_printf(ctx, out, "/>\n");
+					fz_drop_path(ctx, path);
+				}
+				else
+				{
+					rect = fz_empty_rect;
+				}
 			}
-			else
+			else if (fz_font_t3_procs(ctx, span->font))
 			{
-				fz_bound_glyph(ctx, span->font, gid, &fz_identity, &rect);
+				rect = fz_bound_glyph(ctx, span->font, gid, fz_identity);
 				shift.e = -rect.x0;
 				shift.f = -rect.y0;
-				out = start_def(ctx, sdev);
-				fz_write_printf(ctx, out, "<symbol id=\"font_%x_%x\">\n", fnt->id, gid);
-				fz_run_t3_glyph(ctx, span->font, gid, &shift, dev);
+				fz_run_t3_glyph(ctx, span->font, gid, shift, dev);
+				fnt = &sdev->fonts[font_idx]; /* recursion may realloc the font array! */
 			}
 			fz_write_printf(ctx, out, "</symbol>\n");
 			out = end_def(ctx, sdev);
@@ -506,40 +531,38 @@ svg_dev_text_span_as_paths_defs(fz_context *ctx, fz_device *dev, fz_text_span *s
 }
 
 static void
-svg_dev_text_span_as_paths_fill(fz_context *ctx, fz_device *dev, const fz_text_span *span, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha, font *fnt, const fz_color_params *color_params)
+svg_dev_text_span_as_paths_fill(fz_context *ctx, fz_device *dev, const fz_text_span *span, fz_matrix ctm,
+	fz_colorspace *colorspace, const float *color, float alpha, font *fnt, fz_color_params color_params)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out = sdev->out;
-
-	fz_matrix local_trm, local_trm2;
-	int i;
 	fz_matrix shift = { 1, 0, 0, 1, 0, 0};
+	fz_matrix trm, mtx;
+	int i;
 
 	/* Rely on the fact that trm.{e,f} == 0 */
-	local_trm.a = span->trm.a;
-	local_trm.b = span->trm.b;
-	local_trm.c = span->trm.c;
-	local_trm.d = span->trm.d;
-	local_trm.e = 0;
-	local_trm.f = 0;
+	trm.a = span->trm.a;
+	trm.b = span->trm.b;
+	trm.c = span->trm.c;
+	trm.d = span->trm.d;
+	trm.e = 0;
+	trm.f = 0;
 
 	for (i=0; i < span->len; i++)
 	{
 		fz_text_item *it = &span->items[i];
 		int gid = it->gid;
-
 		if (gid < 0)
 			continue;
 
 		shift.e = fnt->sentlist[gid].x_off;
 		shift.f = fnt->sentlist[gid].y_off;
-		local_trm.e = it->x;
-		local_trm.f = it->y;
-		fz_concat(&local_trm2, &local_trm, ctm);
-		fz_concat(&local_trm2, &shift, &local_trm2);
+		trm.e = it->x;
+		trm.f = it->y;
+		mtx = fz_concat(shift, fz_concat(trm, ctm));
+
 		fz_write_printf(ctx, out, "<use xlink:href=\"#font_%x_%x\"", fnt->id, gid);
-		svg_dev_ctm(ctx, sdev, &local_trm2);
+		svg_dev_ctm(ctx, sdev, mtx);
 		svg_dev_fill_color(ctx, sdev, colorspace, color, alpha, color_params);
 		fz_write_printf(ctx, out, "/>\n");
 	}
@@ -547,40 +570,39 @@ svg_dev_text_span_as_paths_fill(fz_context *ctx, fz_device *dev, const fz_text_s
 
 static void
 svg_dev_text_span_as_paths_stroke(fz_context *ctx, fz_device *dev, const fz_text_span *span,
-	const fz_stroke_state *stroke, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha, font *fnt, const fz_color_params *color_params)
+	const fz_stroke_state *stroke, fz_matrix ctm,
+	fz_colorspace *colorspace, const float *color, float alpha, font *fnt, fz_color_params color_params)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out = sdev->out;
-	fz_matrix local_trm, local_trm2;
-	int i;
 	fz_matrix shift = { 1, 0, 0, 1, 0, 0};
+	fz_matrix trm, mtx;
+	int i;
 
 	/* Rely on the fact that trm.{e,f} == 0 */
-	local_trm.a = span->trm.a;
-	local_trm.b = span->trm.b;
-	local_trm.c = span->trm.c;
-	local_trm.d = span->trm.d;
-	local_trm.e = 0;
-	local_trm.f = 0;
+	trm.a = span->trm.a;
+	trm.b = span->trm.b;
+	trm.c = span->trm.c;
+	trm.d = span->trm.d;
+	trm.e = 0;
+	trm.f = 0;
 
 	for (i=0; i < span->len; i++)
 	{
 		fz_text_item *it = &span->items[i];
 		int gid = it->gid;
-
 		if (gid < 0)
 			continue;
 
 		shift.e = fnt->sentlist[gid].x_off;
 		shift.f = fnt->sentlist[gid].y_off;
-		local_trm.e = it->x;
-		local_trm.f = it->y;
-		fz_concat(&local_trm2, &local_trm, ctm);
-		fz_concat(&local_trm2, &shift, &local_trm2);
+		trm.e = it->x;
+		trm.f = it->y;
+		mtx = fz_concat(shift, fz_concat(trm, ctm));
+
 		fz_write_printf(ctx, out, "<use xlink:href=\"#font_%x_%x\"", fnt->id, gid);
-		svg_dev_stroke_state(ctx, sdev, stroke, &local_trm2);
-		svg_dev_ctm(ctx, sdev, &local_trm2);
+		svg_dev_stroke_state(ctx, sdev, stroke, mtx);
+		svg_dev_ctm(ctx, sdev, mtx);
 		svg_dev_stroke_color(ctx, sdev, colorspace, color, alpha, color_params);
 		fz_write_printf(ctx, out, "/>\n");
 	}
@@ -589,8 +611,8 @@ svg_dev_text_span_as_paths_stroke(fz_context *ctx, fz_device *dev, const fz_text
 /* Entry points */
 
 static void
-svg_dev_fill_path(fz_context *ctx, fz_device *dev, const fz_path *path, int even_odd, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+svg_dev_fill_path(fz_context *ctx, fz_device *dev, const fz_path *path, int even_odd, fz_matrix ctm,
+	fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out = sdev->out;
@@ -605,22 +627,22 @@ svg_dev_fill_path(fz_context *ctx, fz_device *dev, const fz_path *path, int even
 }
 
 static void
-svg_dev_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, const fz_stroke_state *stroke, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+svg_dev_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, const fz_stroke_state *stroke, fz_matrix ctm,
+	fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out = sdev->out;
 
 	fz_write_printf(ctx, out, "<path");
 	svg_dev_ctm(ctx, sdev, ctm);
-	svg_dev_stroke_state(ctx, sdev, stroke, &fz_identity);
+	svg_dev_stroke_state(ctx, sdev, stroke, fz_identity);
 	svg_dev_stroke_color(ctx, sdev, colorspace, color, alpha, color_params);
 	svg_dev_path(ctx, sdev, path);
 	fz_write_printf(ctx, out, "/>\n");
 }
 
 static void
-svg_dev_clip_path(fz_context *ctx, fz_device *dev, const fz_path *path, int even_odd, const fz_matrix *ctm, const fz_rect *scissor)
+svg_dev_clip_path(fz_context *ctx, fz_device *dev, const fz_path *path, int even_odd, fz_matrix ctm, fz_rect scissor)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out;
@@ -640,7 +662,7 @@ svg_dev_clip_path(fz_context *ctx, fz_device *dev, const fz_path *path, int even
 }
 
 static void
-svg_dev_clip_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, const fz_stroke_state *stroke, const fz_matrix *ctm, const fz_rect *scissor)
+svg_dev_clip_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, const fz_stroke_state *stroke, fz_matrix ctm, fz_rect scissor)
 {
 	svg_device *sdev = (svg_device*)dev;
 
@@ -649,15 +671,15 @@ svg_dev_clip_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, c
 	int num = sdev->id++;
 	float white[3] = { 1, 1, 1 };
 
-	fz_bound_path(ctx, path, stroke, ctm, &bounds);
+	bounds = fz_bound_path(ctx, path, stroke, ctm);
 
 	out = start_def(ctx, sdev);
 	fz_write_printf(ctx, out, "<mask id=\"ma%d\" x=\"%g\" y=\"%g\" width=\"%g\" height=\"%g\" maskUnits=\"userSpaceOnUse\" maskContentUnits=\"userSpaceOnUse\">\n",
 		num, bounds.x0, bounds.y0, bounds.x1 - bounds.x0, bounds.y1 - bounds.y0);
 	fz_write_printf(ctx, out, "<path");
 	svg_dev_ctm(ctx, sdev, ctm);
-	svg_dev_stroke_state(ctx, sdev, stroke, &fz_identity);
-	svg_dev_stroke_color(ctx, sdev, fz_device_rgb(ctx), white, 1, NULL);
+	svg_dev_stroke_state(ctx, sdev, stroke, fz_identity);
+	svg_dev_stroke_color(ctx, sdev, fz_device_rgb(ctx), white, 1, fz_default_color_params);
 	svg_dev_path(ctx, sdev, path);
 	fz_write_printf(ctx, out, "/>\n</mask>\n");
 	out = end_def(ctx, sdev);
@@ -665,8 +687,8 @@ svg_dev_clip_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, c
 }
 
 static void
-svg_dev_fill_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+svg_dev_fill_text(fz_context *ctx, fz_device *dev, const fz_text *text, fz_matrix ctm,
+	fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out = sdev->out;
@@ -693,8 +715,8 @@ svg_dev_fill_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz
 }
 
 static void
-svg_dev_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+svg_dev_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, fz_matrix ctm,
+	fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out = sdev->out;
@@ -721,7 +743,7 @@ svg_dev_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const
 }
 
 static void
-svg_dev_clip_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm, const fz_rect *scissor)
+svg_dev_clip_text(fz_context *ctx, fz_device *dev, const fz_text *text, fz_matrix ctm, fz_rect scissor)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out = sdev->out;
@@ -732,7 +754,7 @@ svg_dev_clip_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz
 	font *fnt;
 	fz_text_span *span;
 
-	fz_bound_text(ctx, text, NULL, ctm, &bounds);
+	bounds = fz_bound_text(ctx, text, NULL, ctm);
 
 	out = start_def(ctx, sdev);
 	fz_write_printf(ctx, out, "<mask id=\"ma%d\" x=\"%g\" y=\"%g\" width=\"%g\" height=\"%g\"",
@@ -743,7 +765,7 @@ svg_dev_clip_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz
 		for (span = text->head; span; span = span->next)
 		{
 			fz_write_printf(ctx, out, "<text");
-			svg_dev_fill_color(ctx, sdev, fz_device_rgb(ctx), white, 1, NULL);
+			svg_dev_fill_color(ctx, sdev, fz_device_rgb(ctx), white, 1, fz_default_color_params);
 			svg_dev_text_span(ctx, sdev, ctm, span);
 		}
 	}
@@ -752,7 +774,7 @@ svg_dev_clip_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz
 		for (span = text->head; span; span = span->next)
 		{
 			fnt = svg_dev_text_span_as_paths_defs(ctx, dev, span, ctm);
-			svg_dev_text_span_as_paths_fill(ctx, dev, span, ctm, fz_device_rgb(ctx), white, 1.0f, fnt, NULL);
+			svg_dev_text_span_as_paths_fill(ctx, dev, span, ctm, fz_device_rgb(ctx), white, 1.0f, fnt, fz_default_color_params);
 		}
 	}
 	fz_write_printf(ctx, out, "</mask>\n");
@@ -761,7 +783,7 @@ svg_dev_clip_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz
 }
 
 static void
-svg_dev_clip_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, const fz_matrix *ctm, const fz_rect *scissor)
+svg_dev_clip_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, fz_matrix ctm, fz_rect scissor)
 {
 	svg_device *sdev = (svg_device*)dev;
 
@@ -772,7 +794,7 @@ svg_dev_clip_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, c
 	font *fnt;
 	fz_text_span *span;
 
-	fz_bound_text(ctx, text, NULL, ctm, &bounds);
+	bounds = fz_bound_text(ctx, text, NULL, ctm);
 
 	out = start_def(ctx, sdev);
 	fz_write_printf(ctx, out, "<mask id=\"ma%d\" x=\"%g\" y=\"%g\" width=\"%g\" height=\"%g\"",
@@ -783,8 +805,8 @@ svg_dev_clip_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, c
 		for (span = text->head; span; span = span->next)
 		{
 			fz_write_printf(ctx, out, "<text");
-			svg_dev_stroke_state(ctx, sdev, stroke, &fz_identity);
-			svg_dev_stroke_color(ctx, sdev, fz_device_rgb(ctx), white, 1, NULL);
+			svg_dev_stroke_state(ctx, sdev, stroke, fz_identity);
+			svg_dev_stroke_color(ctx, sdev, fz_device_rgb(ctx), white, 1, fz_default_color_params);
 			svg_dev_text_span(ctx, sdev, ctm, span);
 		}
 	}
@@ -793,7 +815,7 @@ svg_dev_clip_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, c
 		for (span = text->head; span; span = span->next)
 		{
 			fnt = svg_dev_text_span_as_paths_defs(ctx, dev, span, ctm);
-			svg_dev_text_span_as_paths_stroke(ctx, dev, span, stroke, ctm, fz_device_rgb(ctx), white, 1.0f, fnt, NULL);
+			svg_dev_text_span_as_paths_stroke(ctx, dev, span, stroke, ctm, fz_device_rgb(ctx), white, 1.0f, fnt, fz_default_color_params);
 		}
 	}
 	fz_write_printf(ctx, out, "</mask>\n");
@@ -802,7 +824,7 @@ svg_dev_clip_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, c
 }
 
 static void
-svg_dev_ignore_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm)
+svg_dev_ignore_text(fz_context *ctx, fz_device *dev, const fz_text *text, fz_matrix ctm)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out = sdev->out;
@@ -815,7 +837,7 @@ svg_dev_ignore_text(fz_context *ctx, fz_device *dev, const fz_text *text, const
 		for (span = text->head; span; span = span->next)
 		{
 			fz_write_printf(ctx, out, "<text");
-			svg_dev_fill_color(ctx, sdev, fz_device_rgb(ctx), black, 0.0f, NULL);
+			svg_dev_fill_color(ctx, sdev, fz_device_rgb(ctx), black, 0.0f, fz_default_color_params);
 			svg_dev_text_span(ctx, sdev, ctm, span);
 		}
 	}
@@ -827,7 +849,7 @@ svg_dev_ignore_text(fz_context *ctx, fz_device *dev, const fz_text *text, const
  * have conniptions. We therefore have an option that is
  * made to avoid this (reuse-images=no). */
 static void
-svg_send_image(fz_context *ctx, svg_device *sdev, fz_image *img, const fz_color_params *color_params)
+svg_send_image(fz_context *ctx, svg_device *sdev, fz_image *img, fz_color_params color_params)
 {
 	fz_output *out = sdev->out;
 	int i;
@@ -851,21 +873,18 @@ svg_send_image(fz_context *ctx, svg_device *sdev, fz_image *img, const fz_color_
 			int new_max = sdev->max_images * 2;
 			if (new_max == 0)
 				new_max = 32;
-			sdev->images = fz_resize_array(ctx, sdev->images, new_max, sizeof(image));
+			sdev->images = fz_realloc_array(ctx, sdev->images, new_max, image);
 			sdev->max_images = new_max;
 		}
 
 		id = sdev->id++;
 		out = start_def(ctx, sdev);
 		fz_write_printf(ctx, out, "<symbol id=\"im%d\" viewBox=\"0 0 %d %d\">\n", id, img->w, img->h);
-	}
 
-	fz_write_printf(ctx, out, "<image width=\"%d\" height=\"%d\" xlink:href=\"data:", img->w, img->h);
-	fz_write_image_as_data_uri(ctx, out, img);
-	fz_write_printf(ctx, out, "\"/>\n");
+		fz_write_printf(ctx, out, "<image width=\"%d\" height=\"%d\" xlink:href=\"", img->w, img->h);
+		fz_write_image_as_data_uri(ctx, out, img);
+		fz_write_printf(ctx, out, "\"/>\n");
 
-	if (sdev->reuse_images)
-	{
 		fz_write_printf(ctx, out, "</symbol>\n");
 		out = end_def(ctx, sdev);
 
@@ -876,66 +895,78 @@ svg_send_image(fz_context *ctx, svg_device *sdev, fz_image *img, const fz_color_
 		fz_write_printf(ctx, out, "<use xlink:href=\"#im%d\" x=\"0\" y=\"0\" width=\"%d\" height=\"%d\"/>\n",
 				id, img->w, img->h);
 	}
+	else
+	{
+		fz_write_printf(ctx, out, "<image width=\"%d\" height=\"%d\" xlink:href=\"", img->w, img->h);
+		fz_write_image_as_data_uri(ctx, out, img);
+		fz_write_printf(ctx, out, "\"/>\n");
+	}
 }
 
 static void
-svg_dev_fill_image(fz_context *ctx, fz_device *dev, fz_image *image, const fz_matrix *ctm, float alpha, const fz_color_params *color_params)
+svg_dev_fill_image(fz_context *ctx, fz_device *dev, fz_image *image, fz_matrix ctm, float alpha, fz_color_params color_params)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out = sdev->out;
 
-	fz_matrix local_ctm = *ctm;
+	fz_matrix local_ctm = ctm;
 	fz_matrix scale = { 0 };
 
+	if (alpha == 0)
+		return;
+
 	scale.a = 1.0f / image->w;
 	scale.d = 1.0f / image->h;
 
-	fz_concat(&local_ctm, &scale, ctm);
+	local_ctm = fz_concat(scale, ctm);
 	fz_write_printf(ctx, out, "<g");
 	if (alpha != 1.0f)
 		fz_write_printf(ctx, out, " opacity=\"%g\"", alpha);
-	svg_dev_ctm(ctx, sdev, &local_ctm);
+	svg_dev_ctm(ctx, sdev, local_ctm);
 	fz_write_printf(ctx, out, ">\n");
 	svg_send_image(ctx, sdev, image, color_params);
 	fz_write_printf(ctx, out, "</g>\n");
 }
 
 static void
-svg_dev_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, const fz_matrix *ctm, float alpha, const fz_color_params *color_params)
+svg_dev_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, fz_matrix ctm, float alpha, fz_color_params color_params)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out = sdev->out;
-	fz_rect rect;
 	fz_irect bbox;
 	fz_pixmap *pix;
-	fz_buffer *buf = NULL;
+	fz_rect scissor = fz_device_current_scissor(ctx, dev);
 
-	fz_var(buf);
-
-	if (dev->container_len == 0)
+	if (alpha == 0)
 		return;
 
-	fz_round_rect(&bbox, fz_intersect_rect(fz_bound_shade(ctx, shade, ctm, &rect), &dev->container[dev->container_len-1].scissor));
-	if (fz_is_empty_irect(&bbox))
+	if (fz_is_infinite_rect(scissor))
+	{
+		scissor.x0 = 0;
+		scissor.x1 = sdev->page_width;
+		scissor.y0 = 0;
+		scissor.y1 = sdev->page_height;
+	}
+
+	bbox = fz_round_rect(fz_intersect_rect(fz_bound_shade(ctx, shade, ctm), scissor));
+	if (fz_is_empty_irect(bbox))
 		return;
-	pix = fz_new_pixmap_with_bbox(ctx, fz_device_rgb(ctx), &bbox, NULL, 1);
+	pix = fz_new_pixmap_with_bbox(ctx, fz_device_rgb(ctx), bbox, NULL, 1);
 	fz_clear_pixmap(ctx, pix);
 
 	fz_try(ctx)
 	{
-		fz_paint_shade(ctx, shade, NULL, ctm, pix, color_params, &bbox, NULL);
-		buf = fz_new_buffer_from_pixmap_as_png(ctx, pix, color_params);
+		fz_paint_shade(ctx, shade, NULL, ctm, pix, color_params, bbox, NULL, NULL);
 		if (alpha != 1.0f)
 			fz_write_printf(ctx, out, "<g opacity=\"%g\">\n", alpha);
-		fz_write_printf(ctx, out, "<image x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" xlink:href=\"data:image/png;base64,", pix->x, pix->y, pix->w, pix->h);
-		fz_write_base64_buffer(ctx, out, buf, 1);
+		fz_write_printf(ctx, out, "<image x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" xlink:href=\"", pix->x, pix->y, pix->w, pix->h);
+		fz_write_pixmap_as_data_uri(ctx, out, pix);
 		fz_write_printf(ctx, out, "\"/>\n");
 		if (alpha != 1.0f)
 			fz_write_printf(ctx, out, "</g>\n");
 	}
 	fz_always(ctx)
 	{
-		fz_drop_buffer(ctx, buf);
 		fz_drop_pixmap(ctx, pix);
 	}
 	fz_catch(ctx)
@@ -945,19 +976,19 @@ svg_dev_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, const fz_ma
 }
 
 static void
-svg_dev_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+svg_dev_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, fz_matrix ctm,
+	fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out;
-	fz_matrix local_ctm = *ctm;
+	fz_matrix local_ctm = ctm;
 	fz_matrix scale = { 0 };
 	int mask = sdev->id++;
 
 	scale.a = 1.0f / image->w;
 	scale.d = 1.0f / image->h;
 
-	fz_concat(&local_ctm, &scale, ctm);
+	local_ctm = fz_concat(scale, ctm);
 	out = start_def(ctx, sdev);
 	fz_write_printf(ctx, out, "<mask id=\"ma%d\">\n", mask);
 	svg_send_image(ctx, sdev, image, color_params);
@@ -965,28 +996,28 @@ svg_dev_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, const
 	out = end_def(ctx, sdev);
 	fz_write_printf(ctx, out, "<rect x=\"0\" y=\"0\" width=\"%d\" height=\"%d\"", image->w, image->h);
 	svg_dev_fill_color(ctx, sdev, colorspace, color, alpha, color_params);
-	svg_dev_ctm(ctx, sdev, &local_ctm);
+	svg_dev_ctm(ctx, sdev, local_ctm);
 	fz_write_printf(ctx, out, " mask=\"url(#ma%d)\"/>\n", mask);
 }
 
 static void
-svg_dev_clip_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, const fz_matrix *ctm, const fz_rect *scissor)
+svg_dev_clip_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, fz_matrix ctm, fz_rect scissor)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out;
-	fz_matrix local_ctm = *ctm;
+	fz_matrix local_ctm = ctm;
 	fz_matrix scale = { 0 };
 	int mask = sdev->id++;
 
 	scale.a = 1.0f / image->w;
 	scale.d = 1.0f / image->h;
 
-	fz_concat(&local_ctm, &scale, ctm);
+	local_ctm = fz_concat(scale, ctm);
 	out = start_def(ctx, sdev);
 	fz_write_printf(ctx, out, "<mask id=\"ma%d\">\n<g", mask);
-	svg_dev_ctm(ctx, sdev, &local_ctm);
+	svg_dev_ctm(ctx, sdev, local_ctm);
 	fz_write_printf(ctx, out, ">\n");
-	svg_send_image(ctx, sdev, image, NULL/* FIXME */);
+	svg_send_image(ctx, sdev, image, fz_default_color_params/* FIXME */);
 	fz_write_printf(ctx, out, "</g>\n</mask>\n");
 	out = end_def(ctx, sdev);
 	fz_write_printf(ctx, out, "<g mask=\"url(#ma%d)\">\n", mask);
@@ -1003,7 +1034,7 @@ svg_dev_pop_clip(fz_context *ctx, fz_device *dev)
 }
 
 static void
-svg_dev_begin_mask(fz_context *ctx, fz_device *dev, const fz_rect *bbox, int luminosity, fz_colorspace *colorspace, const float *color, const fz_color_params *color_params)
+svg_dev_begin_mask(fz_context *ctx, fz_device *dev, fz_rect bbox, int luminosity, fz_colorspace *colorspace, const float *color, fz_color_params color_params)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out;
@@ -1024,7 +1055,7 @@ svg_dev_end_mask(fz_context *ctx, fz_device *dev)
 	int mask = 0;
 
 	if (dev->container_len > 0)
-		mask = (int)dev->container[dev->container_len-1].user;
+		mask = dev->container[dev->container_len-1].user;
 
 	fz_write_printf(ctx, out, "\"/>\n</mask>\n");
 	out = end_def(ctx, sdev);
@@ -1032,16 +1063,54 @@ svg_dev_end_mask(fz_context *ctx, fz_device *dev)
 }
 
 static void
-svg_dev_begin_group(fz_context *ctx, fz_device *dev, const fz_rect *bbox, fz_colorspace *cs, int isolated, int knockout, int blendmode, float alpha)
+svg_dev_begin_group(fz_context *ctx, fz_device *dev, fz_rect bbox, fz_colorspace *cs, int isolated, int knockout, int blendmode, float alpha)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out = sdev->out;
 
+	/* SVG only supports normal/multiply/screen/darken/lighten,
+	 * but we'll send them all, as the spec says that unrecognised
+	 * ones are treated as normal. */
+	static char *blend_names[] = {
+		"normal",	/* FZ_BLEND_NORMAL */
+		"multiply",	/* FZ_BLEND_MULTIPLY */
+		"screen",	/* FZ_BLEND_SCREEN */
+		"overlay",	/* FZ_BLEND_OVERLAY */
+		"darken",	/* FZ_BLEND_DARKEN */
+		"lighten",	/* FZ_BLEND_LIGHTEN */
+		"color_dodge",	/* FZ_BLEND_COLOR_DODGE */
+		"color_burn",	/* FZ_BLEND_COLOR_BURN */
+		"hard_light",	/* FZ_BLEND_HARD_LIGHT */
+		"soft_light",	/* FZ_BLEND_SOFT_LIGHT */
+		"difference",	/* FZ_BLEND_DIFFERENCE */
+		"exclusion",	/* FZ_BLEND_EXCLUSION */
+		"hue",		/* FZ_BLEND_HUE */
+		"saturation",	/* FZ_BLEND_SATURATION */
+		"color",	/* FZ_BLEND_COLOR */
+		"luminosity",	/* FZ_BLEND_LUMINOSITY */
+	};
+
+	if (blendmode < FZ_BLEND_NORMAL || blendmode > FZ_BLEND_LUMINOSITY)
+		blendmode = FZ_BLEND_NORMAL;
+	if (blendmode != FZ_BLEND_NORMAL && (sdev->blend_bitmask & (1<<blendmode)) == 0)
+	{
+		sdev->blend_bitmask |= (1<<blendmode);
+		out = start_def(ctx, sdev);
+		fz_write_printf(ctx, out,
+				"<filter id=\"blend_%d\"><feBlend mode=\"%s\" in2=\"BackgroundImage\" in=\"SourceGraphic\"/></filter>\n",
+				blendmode, blend_names[blendmode]);
+		out = end_def(ctx, sdev);
+	}
+
+	/* FIXME: Handle alpha == 0 somehow? */
 	/* SVG 1.1 doesn't support adequate blendmodes/knockout etc, so just ignore it for now */
 	if (alpha == 1)
-		fz_write_printf(ctx, out, "<g>\n");
+		fz_write_printf(ctx, out, "<g");
 	else
-		fz_write_printf(ctx, out, "<g opacity=\"%g\">\n", alpha);
+		fz_write_printf(ctx, out, "<g opacity=\"%g\"", alpha);
+	if (blendmode != FZ_BLEND_NORMAL)
+		fz_write_printf(ctx, out, " filter=\"url(#blend_%d)\"", blendmode);
+	fz_write_printf(ctx, out, ">\n");
 }
 
 static void
@@ -1054,7 +1123,7 @@ svg_dev_end_group(fz_context *ctx, fz_device *dev)
 }
 
 static int
-svg_dev_begin_tile(fz_context *ctx, fz_device *dev, const fz_rect *area, const fz_rect *view, float xstep, float ystep, const fz_matrix *ctm, int id)
+svg_dev_begin_tile(fz_context *ctx, fz_device *dev, fz_rect area, fz_rect view, float xstep, float ystep, fz_matrix ctm, int id)
 {
 	svg_device *sdev = (svg_device*)dev;
 	fz_output *out;
@@ -1065,15 +1134,26 @@ svg_dev_begin_tile(fz_context *ctx, fz_device *dev, const fz_rect *area, const f
 	{
 		int n = (sdev->num_tiles == 0 ? 4 : sdev->num_tiles * 2);
 
-		sdev->tiles = fz_resize_array(ctx, sdev->tiles, n, sizeof(tile));
+		sdev->tiles = fz_realloc_array(ctx, sdev->tiles, n, tile);
 		sdev->max_tiles = n;
 	}
 	num = sdev->num_tiles++;
 	t = &sdev->tiles[num];
-	t->area = *area;
-	t->view = *view;
-	t->ctm = *ctm;
+	t->area = area;
+	t->view = view;
+	t->ctm = ctm;
 	t->pattern = sdev->id++;
+
+	xstep = fabsf(xstep);
+	ystep = fabsf(ystep);
+	if (xstep == 0 || ystep == 0) {
+		fz_warn(ctx, "Pattern cannot have x or ystep == 0.");
+		if (xstep == 0)
+			xstep = 1;
+		if (ystep == 0)
+			ystep = 1;
+	}
+
 	t->step.x = xstep;
 	t->step.y = ystep;
 
@@ -1089,7 +1169,7 @@ svg_dev_begin_tile(fz_context *ctx, fz_device *dev, const fz_rect *area, const f
 	/* The first thing we do is to capture the contents of the pattern
 	 * as a symbol we can reuse. */
 	out = start_def(ctx, sdev);
-	fz_write_printf(ctx, out, "<symbol id=\"pac%d\">\n", t->pattern);
+	fz_write_printf(ctx, out, "<symbol id=\"pac%d\" style=\"overflow:visible\">\n", t->pattern);
 
 	return 0;
 }
@@ -1139,9 +1219,9 @@ svg_dev_end_tile(fz_context *ctx, fz_device *dev)
 
 	/* All the pattern contents will have their own ctm applied. Let's
 	 * undo the current one to allow for this */
-	fz_invert_matrix(&inverse, &t->ctm);
+	inverse = fz_invert_matrix(t->ctm);
 	fz_write_printf(ctx, out, "<g");
-	svg_dev_ctm(ctx, sdev, &inverse);
+	svg_dev_ctm(ctx, sdev, inverse);
 	fz_write_printf(ctx, out, ">\n");
 
 	w = t->view.x1 - t->view.x0;
@@ -1159,7 +1239,7 @@ svg_dev_end_tile(fz_context *ctx, fz_device *dev)
 
 	/* Finally, fill a rectangle with the pattern. */
 	fz_write_printf(ctx, out, "<rect");
-	svg_dev_ctm(ctx, sdev, &t->ctm);
+	svg_dev_ctm(ctx, sdev, t->ctm);
 	fz_write_printf(ctx, out, " fill=\"url(#pa%d)\" x=\"%g\" y=\"%g\" width=\"%g\" height=\"%g\"/>\n",
 		t->pattern, t->area.x0, t->area.y0, t->area.x1 - t->area.x0, t->area.y1 - t->area.y0);
 }
@@ -1199,6 +1279,10 @@ svg_dev_close_device(fz_context *ctx, fz_device *dev)
 		sdev->layers--;
 	}
 
+	if (sdev->save_id)
+		*sdev->save_id = sdev->id;
+
+	fz_write_printf(ctx, out, "</g>\n");
 	fz_write_printf(ctx, out, "</svg>\n");
 }
 
@@ -1224,7 +1308,7 @@ svg_dev_drop_device(fz_context *ctx, fz_device *dev)
 	fz_free(ctx, sdev->images);
 }
 
-fz_device *fz_new_svg_device(fz_context *ctx, fz_output *out, float page_width, float page_height, int text_format, int reuse_images)
+fz_device *fz_new_svg_device_with_id(fz_context *ctx, fz_output *out, float page_width, float page_height, int text_format, int reuse_images, int *id)
 {
 	svg_device *dev = fz_new_derived_device(ctx, svg_device);
 
@@ -1260,14 +1344,15 @@ fz_device *fz_new_svg_device(fz_context *ctx, fz_output *out, float page_width,
 	dev->super.begin_layer = svg_dev_begin_layer;
 	dev->super.end_layer = svg_dev_end_layer;
 
-	dev->super.hints |= FZ_MAINTAIN_CONTAINER_STACK;
-
 	dev->out = out;
 	dev->out_store = out;
-	dev->id = 0;
+	dev->save_id = id;
+	dev->id = id ? *id : 0;
 	dev->layers = 0;
 	dev->text_as_text = (text_format == FZ_SVG_TEXT_AS_TEXT);
 	dev->reuse_images = reuse_images;
+	dev->page_width = page_width;
+	dev->page_height = page_height;
 
 	fz_write_printf(ctx, out, "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n");
 	fz_write_printf(ctx, out, "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n");
@@ -1275,6 +1360,12 @@ fz_device *fz_new_svg_device(fz_context *ctx, fz_output *out, float page_width,
 		"xmlns:xlink=\"http://www.w3.org/1999/xlink\" version=\"1.1\" "
 		"width=\"%gpt\" height=\"%gpt\" viewBox=\"0 0 %g %g\">\n",
 		page_width, page_height, page_width, page_height);
+	fz_write_printf(ctx, out, "<g enable-background=\"new\">\n");
 
 	return (fz_device*)dev;
 }
+
+fz_device *fz_new_svg_device(fz_context *ctx, fz_output *out, float page_width, float page_height, int text_format, int reuse_images)
+{
+	return fz_new_svg_device_with_id(ctx, out, page_width, page_height, text_format, reuse_images, NULL);
+}
