@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 /**
  * Mu Office Library
  *
@@ -13,6 +35,8 @@
 #include "mupdf/helpers/mu-threads.h"
 #include "mupdf/memento.h"
 
+#include <assert.h>
+
 enum
 {
 	MuError_OK = 0,
@@ -104,7 +128,7 @@ static void muoffice_lock(void *user, int lock);
 
 static void muoffice_unlock(void *user, int lock);
 
-struct MuOfficeLib_s
+struct MuOfficeLib
 {
 	fz_context *ctx;
 	mu_mutex mutexes[FZ_LOCK_MAX+1];
@@ -136,6 +160,16 @@ static void muoffice_unlock(void *user, int lock)
 	mu_unlock_mutex(&mu->mutexes[lock]);
 }
 
+static void muoffice_doc_lock(MuOfficeLib *mu)
+{
+	mu_lock_mutex(&mu->mutexes[DOCLOCK]);
+}
+
+static void muoffice_doc_unlock(MuOfficeLib *mu)
+{
+	mu_unlock_mutex(&mu->mutexes[DOCLOCK]);
+}
+
 static void fin_muoffice_locks(MuOfficeLib *mu)
 {
 	int i;
@@ -282,7 +316,7 @@ char * MuOfficeLib_getSupportedFileExtensions(void)
 	return NULL;
 }
 
-struct MuOfficeDoc_s
+struct MuOfficeDoc
 {
 	MuOfficeLib *mu;
 	fz_context *ctx;
@@ -300,7 +334,7 @@ struct MuOfficeDoc_s
 	MuOfficePage *pages;
 };
 
-struct MuOfficePage_s
+struct MuOfficePage
 {
 	MuOfficePage *next;
 	MuOfficeDoc *doc;
@@ -311,7 +345,7 @@ struct MuOfficePage_s
 	fz_display_list *list;
 };
 
-struct MuOfficeRender_s
+struct MuOfficeRender
 {
 	MuOfficePage *page;
 	float zoom;
@@ -337,7 +371,7 @@ static void load_worker(void *arg)
 		return;
 	}
 
-	fz_lock(ctx, DOCLOCK);
+	muoffice_doc_lock(doc->mu);
 
 	fz_try(ctx)
 	{
@@ -375,7 +409,7 @@ static void load_worker(void *arg)
 		err = MuOfficeDocErrorType_UnableToLoadDocument;
 
 fail:
-	fz_unlock(ctx, DOCLOCK);
+	muoffice_doc_unlock(doc->mu);
 
 	if (err)
 		doc->error(doc->cookie, err);
@@ -670,7 +704,7 @@ MuError MuOfficeDoc_getPage(	MuOfficeDoc          *doc,
 	if (page == NULL)
 		return MuError_OOM;
 
-	fz_lock(ctx, DOCLOCK);
+	muoffice_doc_lock(doc->mu);
 
 	fz_try(ctx)
 	{
@@ -689,7 +723,7 @@ MuError MuOfficeDoc_getPage(	MuOfficeDoc          *doc,
 		err = MuError_Generic;
 	}
 
-	fz_unlock(ctx, DOCLOCK);
+	muoffice_doc_unlock(doc->mu);
 
 	return err;
 }
@@ -734,14 +768,14 @@ MuError MuOfficeDoc_run(MuOfficeDoc *doc, void (*fn)(fz_context *ctx, fz_documen
 	if (ctx == NULL)
 		return MuError_OOM;
 
-	fz_lock(ctx, DOCLOCK);
+	muoffice_doc_lock(doc->mu);
 
 	fz_try(ctx)
 		fn(ctx, doc->doc, arg);
 	fz_catch(ctx)
 		err = MuError_Generic;
 
-	fz_unlock(ctx, DOCLOCK);
+	muoffice_doc_unlock(doc->mu);
 
 	fz_drop_context(ctx);
 
@@ -805,7 +839,7 @@ MuError MuOfficePage_getSize(	MuOfficePage *page,
 	if (!doc)
 		return MuError_BadNull;
 
-	fz_bound_page(doc->ctx, page->page, &rect);
+	rect = fz_bound_page(doc->ctx, page->page);
 
 	/* MuPDF measures in points (72ths of an inch). This API wants
 	 * 90ths of an inch, so adjust. */
@@ -846,7 +880,7 @@ MuError MuOfficePage_calculateZoom(	MuOfficePage *page,
 	if (!doc)
 		return MuError_BadNull;
 
-	fz_bound_page(doc->ctx, page->page, &rect);
+	rect = fz_bound_page(doc->ctx, page->page);
 
 	/* MuPDF measures in points (72ths of an inch). This API wants
 	 * 90ths of an inch, so adjust. */
@@ -891,7 +925,7 @@ MuError MuOfficePage_getSizeForZoom(	MuOfficePage *page,
 	if (!doc)
 		return MuError_BadNull;
 
-	fz_bound_page(doc->ctx, page->page, &rect);
+	rect = fz_bound_page(doc->ctx, page->page);
 
 	/* MuPDF measures in points (72ths of an inch). This API wants
 	 * 90ths of an inch, so adjust. */
@@ -899,9 +933,9 @@ MuError MuOfficePage_getSizeForZoom(	MuOfficePage *page,
 	h = 90 * (rect.y1 - rect.y0) / 72;
 
 	if (pWidth)
-		*pWidth = (int)(w * zoom + 0.5);
+		*pWidth = (int)(w * zoom + 0.5f);
 	if (pHeight)
-		*pHeight = (int)(h * zoom + 0.5);
+		*pHeight = (int)(h * zoom + 0.5f);
 
 	return MuError_OK;
 }
@@ -944,14 +978,14 @@ MuError MuOfficePage_run(MuOfficePage *page, void (*fn)(fz_context *ctx, fz_page
 	if (ctx == NULL)
 		return MuError_OOM;
 
-	fz_lock(ctx, DOCLOCK);
+	muoffice_doc_lock(page->doc->mu);
 
 	fz_try(ctx)
 		fn(ctx, page->page, arg);
 	fz_catch(ctx)
 		err = MuError_Generic;
 
-	fz_unlock(ctx, DOCLOCK);
+	muoffice_doc_unlock(page->doc->mu);
 
 	fz_drop_context(ctx);
 
@@ -968,7 +1002,6 @@ static void render_worker(void *arg)
 	fz_device *dev = NULL;
 	float scalex;
 	float scaley;
-	fz_matrix matrix;
 	fz_rect page_bounds;
 	int locked = 0;
 
@@ -983,11 +1016,11 @@ static void render_worker(void *arg)
 	{
 		if (page->list == NULL)
 		{
-			fz_lock(ctx, DOCLOCK);
+			muoffice_doc_lock(page->doc->mu);
 			locked = 1;
 			page->list = fz_new_display_list_from_page(ctx, page->page);
 			locked = 0;
-			fz_unlock(ctx, DOCLOCK);
+			muoffice_doc_unlock(page->doc->mu);
 		}
 		/* Make a pixmap from the bitmap */
 		if (!render->area_valid)
@@ -1001,6 +1034,7 @@ static void render_worker(void *arg)
 						fz_device_rgb(ctx),
 						render->area.renderArea.width,
 						render->area.renderArea.height,
+						NULL,
 						1,
 						render->bitmap->lineSkip,
 						((unsigned char *)render->bitmap->memptr) +
@@ -1009,16 +1043,16 @@ static void render_worker(void *arg)
 		/* Be a bit clever with the scaling to make sure we get
 		 * integer width/heights. First calculate the target
 		 * width/height. */
-		fz_bound_page(ctx, render->page->page, &page_bounds);
-		scalex = (int)(90 * render->zoom * (page_bounds.x1 - page_bounds.x0) / 72 + 0.5);
-		scaley = (int)(90 * render->zoom * (page_bounds.y1 - page_bounds.y0) / 72 + 0.5);
+		page_bounds = fz_bound_page(ctx, render->page->page);
+		scalex = (int)(90 * render->zoom * (page_bounds.x1 - page_bounds.x0) / 72 + 0.5f);
+		scaley = (int)(90 * render->zoom * (page_bounds.y1 - page_bounds.y0) / 72 + 0.5f);
 		/* Now calculate the actual scale factors required */
 		scalex /= (page_bounds.x1 - page_bounds.x0);
 		scaley /= (page_bounds.y1 - page_bounds.y0);
 		/* Render the list */
 		fz_clear_pixmap_with_value(ctx, pixmap, 0xFF);
-		dev = fz_new_draw_device(ctx, fz_post_scale(fz_translate(&matrix, -page_bounds.x0, -page_bounds.y0), scalex, scaley), pixmap);
-		fz_run_display_list(ctx, page->list, dev, &fz_identity, NULL, NULL);
+		dev = fz_new_draw_device(ctx, fz_post_scale(fz_translate(-page_bounds.x0, -page_bounds.y0), scalex, scaley), pixmap);
+		fz_run_display_list(ctx, page->list, dev, fz_identity, fz_infinite_rect, NULL);
 		fz_close_device(ctx, dev);
 	}
 	fz_always(ctx)
@@ -1029,7 +1063,7 @@ static void render_worker(void *arg)
 	fz_catch(ctx)
 	{
 		if (locked)
-			fz_unlock(ctx, DOCLOCK);
+			muoffice_doc_unlock(page->doc->mu);
 		err = MuError_Generic;
 		goto fail;
 	}
