@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 #include "html-imp.h"
 
@@ -58,7 +80,6 @@ struct epub_chapter
 struct epub_page
 {
 	fz_page super;
-	epub_document *doc;
 	epub_chapter *ch;
 	int number;
 	fz_html *html;
@@ -112,8 +133,8 @@ static int count_chapter_pages(fz_context *ctx, epub_document *doc, epub_chapter
 	return acc->pages_in_chapter[ch->number];
 }
 
-static fz_location
-epub_resolve_link(fz_context *ctx, fz_document *doc_, const char *dest, float *xp, float *yp)
+static fz_link_dest
+epub_resolve_link(fz_context *ctx, fz_document *doc_, const char *dest)
 {
 	epub_document *doc = (epub_document*)doc_;
 	epub_chapter *ch;
@@ -140,16 +161,15 @@ epub_resolve_link(fz_context *ctx, fz_document *doc_, const char *dest, float *x
 				if (y >= 0)
 				{
 					int page = y / ph;
-					if (yp) *yp = y - page * ph;
-					return fz_make_location(i, page);
+					return fz_make_link_dest_xyz(i, page, 0, y - page * ph, 0);
 				}
-				return fz_make_location(-1, -1);
+				return fz_make_link_dest_none();
 			}
-			return fz_make_location(i, 0);
+			return fz_make_link_dest_xyz(i, 0, 0, 0, 0);
 		}
 	}
 
-	return fz_make_location(-1, -1);
+	return fz_make_link_dest_none();
 }
 
 static void
@@ -344,7 +364,6 @@ static void
 epub_drop_page(fz_context *ctx, fz_page *page_)
 {
 	epub_page *page = (epub_page *)page_;
-	fz_drop_document(ctx, &page->doc->super);
 	fz_drop_html(ctx, page->html);
 }
 
@@ -418,10 +437,11 @@ epub_get_laid_out_html(fz_context *ctx, epub_document *doc, epub_chapter *ch)
 static fz_rect
 epub_bound_page(fz_context *ctx, fz_page *page_)
 {
+	epub_document *doc = (epub_document*)page_->doc;
 	epub_page *page = (epub_page*)page_;
 	epub_chapter *ch = page->ch;
 	fz_rect bbox;
-	fz_html *html = epub_get_laid_out_html(ctx, page->doc, ch);
+	fz_html *html = epub_get_laid_out_html(ctx, doc, ch);
 
 	bbox.x0 = 0;
 	bbox.y0 = 0;
@@ -443,10 +463,9 @@ static fz_link *
 epub_load_links(fz_context *ctx, fz_page *page_)
 {
 	epub_page *page = (epub_page*)page_;
-	epub_document *doc = page->doc;
 	epub_chapter *ch = page->ch;
 
-	return fz_load_html_links(ctx, page->html, page->number, ch->path, doc);
+	return fz_load_html_links(ctx, page->html, page->number, ch->path);
 }
 
 static fz_bookmark
@@ -498,12 +517,11 @@ epub_load_page(fz_context *ctx, fz_document *doc_, int chapter, int number)
 	{
 		if (i == chapter)
 		{
-			epub_page *page = fz_new_derived_page(ctx, epub_page);
+			epub_page *page = fz_new_derived_page(ctx, epub_page, doc_);
 			page->super.bound_page = epub_bound_page;
 			page->super.run_page_contents = epub_run_page;
 			page->super.load_links = epub_load_links;
 			page->super.drop_page = epub_drop_page;
-			page->doc = (epub_document *)fz_keep_document(ctx, doc_);
 			page->ch = ch;
 			page->number = number;
 			page->html = epub_get_laid_out_html(ctx, doc, ch);
@@ -606,7 +624,7 @@ epub_parse_ncx_imp(fz_context *ctx, epub_document *doc, fz_xml *node, char *base
 				tailp = &(*tailp)->next;
 				outline->title = Memento_label(fz_strdup(ctx, text), "outline_title");
 				outline->uri = Memento_label(fz_strdup(ctx, path), "outline_uri");
-				outline->page = -1;
+				outline->page = fz_make_location(-1, -1);
 				outline->down = epub_parse_ncx_imp(ctx, doc, node, base_uri);
 				outline->is_open = 1;
 			}
@@ -818,6 +836,7 @@ epub_init(fz_context *ctx, fz_archive *zip, fz_stream *accel)
 	epub_document *doc = NULL;
 
 	fz_var(doc);
+	fz_var(zip);
 
 	fz_try(ctx)
 	{
@@ -828,7 +847,7 @@ epub_init(fz_context *ctx, fz_archive *zip, fz_stream *accel)
 		doc->super.drop_document = epub_drop_document;
 		doc->super.layout = epub_layout;
 		doc->super.load_outline = epub_load_outline;
-		doc->super.resolve_link = epub_resolve_link;
+		doc->super.resolve_link_dest = epub_resolve_link;
 		doc->super.make_bookmark = epub_make_bookmark;
 		doc->super.lookup_bookmark = epub_lookup_bookmark;
 		doc->super.count_chapters = epub_count_chapters;
