@@ -1,4 +1,29 @@
-#include "mupdf/html.h"
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
+#include "mupdf/fitz.h"
+#include "html-imp.h"
+
+#include <string.h>
 
 static fz_font *
 fz_load_html_default_font(fz_context *ctx, fz_html_font_set *set, const char *family, int is_bold, int is_italic)
@@ -10,7 +35,7 @@ fz_load_html_default_font(fz_context *ctx, fz_html_font_set *set, const char *fa
 	int idx = (is_mono ? 8 : is_sans ? 4 : 0) + is_bold * 2 + is_italic;
 	if (!set->fonts[idx])
 	{
-		const char *data;
+		const unsigned char *data;
 		int size;
 
 		data = fz_lookup_builtin_font(ctx, real_family, is_bold, is_italic, &size);
@@ -26,8 +51,8 @@ fz_load_html_default_font(fz_context *ctx, fz_html_font_set *set, const char *fa
 
 void
 fz_add_html_font_face(fz_context *ctx, fz_html_font_set *set,
-	const char *family, int is_bold, int is_italic, const char *src,
-	fz_font *font)
+	const char *family, int is_bold, int is_italic, int is_small_caps,
+	const char *src, fz_font *font)
 {
 	fz_html_font_face *custom = fz_malloc_struct(ctx, fz_html_font_face);
 	custom->font = fz_keep_font(ctx, font);
@@ -35,26 +60,38 @@ fz_add_html_font_face(fz_context *ctx, fz_html_font_set *set,
 	custom->family = fz_strdup(ctx, family);
 	custom->is_bold = is_bold;
 	custom->is_italic = is_italic;
+	custom->is_small_caps = is_small_caps;
 	custom->next = set->custom;
 	set->custom = custom;
 }
 
 fz_font *
-fz_load_html_font(fz_context *ctx, fz_html_font_set *set, const char *family, int is_bold, int is_italic)
+fz_load_html_font(fz_context *ctx, fz_html_font_set *set,
+	const char *family, int is_bold, int is_italic, int is_small_caps)
 {
 	fz_html_font_face *custom;
-	const char *data;
+	const unsigned char *data;
+	int best_score = 0;
+	fz_font *best_font = NULL;
 	int size;
 
 	for (custom = set->custom; custom; custom = custom->next)
 	{
-		if (!strcmp(family, custom->family) &&
-				is_bold == custom->is_bold &&
-				is_italic == custom->is_italic)
+		if (!strcmp(family, custom->family))
 		{
-			return custom->font;
+			int score =
+				1 * (is_bold == custom->is_bold) +
+				2 * (is_italic == custom->is_italic) +
+				4 * (is_small_caps == custom->is_small_caps);
+			if (score > best_score)
+			{
+				best_score = score;
+				best_font = custom->font;
+			}
 		}
 	}
+	if (best_font)
+		return best_font;
 
 	data = fz_lookup_builtin_font(ctx, family, is_bold, is_italic, &size);
 	if (data)
@@ -65,7 +102,7 @@ fz_load_html_font(fz_context *ctx, fz_html_font_set *set, const char *family, in
 			flags->fake_bold = 1;
 		if (is_italic && !flags->is_italic)
 			flags->fake_italic = 1;
-		fz_add_html_font_face(ctx, set, family, is_bold, is_italic, "<builtin>", font);
+		fz_add_html_font_face(ctx, set, family, is_bold, is_italic, 0, "<builtin>", font);
 		fz_drop_font(ctx, font);
 		return font;
 	}
@@ -100,7 +137,7 @@ void fz_drop_html_font_set(fz_context *ctx, fz_html_font_set *set)
 		font = next;
 	}
 
-	for (i = 0; i < nelem(set->fonts); ++i)
+	for (i = 0; i < (int)nelem(set->fonts); ++i)
 		fz_drop_font(ctx, set->fonts[i]);
 
 	fz_free(ctx, set);
