@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2022 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -518,21 +518,24 @@ pdf_create_link(fz_context *ctx, pdf_page *page, fz_rect bbox, const char *uri)
 	fz_link **linkp;
 	fz_rect page_mediabox;
 	fz_matrix page_ctm;
+	fz_rect rect;
 
 	fz_var(link);
 	fz_var(ind_obj);
 	fz_var(bs);
 	fz_var(a);
 
-	pdf_page_transform(ctx, page, &page_mediabox, &page_ctm);
-	page_ctm = fz_invert_matrix(page_ctm);
-	bbox = fz_transform_rect(bbox, page_ctm);
+	pdf_begin_operation(ctx, page->doc, "Create Link");
 
 	fz_try(ctx)
 	{
 		int ind_obj_num;
 		pdf_obj *annot_arr;
 
+		pdf_page_transform(ctx, page, &page_mediabox, &page_ctm);
+		page_ctm = fz_invert_matrix(page_ctm);
+		rect = fz_transform_rect(bbox, page_ctm);
+
 		annot_arr = pdf_dict_get(ctx, page->obj, PDF_NAME(Annots));
 		if (annot_arr == NULL)
 		{
@@ -542,19 +545,15 @@ pdf_create_link(fz_context *ctx, pdf_page *page, fz_rect bbox, const char *uri)
 
 		pdf_dict_put(ctx, annot_obj, PDF_NAME(Type), PDF_NAME(Annot));
 		pdf_dict_put(ctx, annot_obj, PDF_NAME(Subtype), PDF_NAME(Link));
-		pdf_dict_put_rect(ctx, annot_obj, PDF_NAME(Rect), bbox);
+		pdf_dict_put_rect(ctx, annot_obj, PDF_NAME(Rect), rect);
 		bs = pdf_new_dict(ctx, doc, 4);
 		pdf_dict_put(ctx, bs, PDF_NAME(S), PDF_NAME(S));
 		pdf_dict_put(ctx, bs, PDF_NAME(Type), PDF_NAME(Border));
 		pdf_dict_put_int(ctx, bs, PDF_NAME(W), 0);
 		pdf_dict_put(ctx, annot_obj, PDF_NAME(BS), bs);
-		if (uri)
-		{
-			a = pdf_new_dict(ctx, doc, 2);
-			pdf_dict_put(ctx, a, PDF_NAME(S), PDF_NAME(URI));
-			pdf_dict_put_text_string(ctx, a, PDF_NAME(URI), uri);
-			pdf_dict_put(ctx, annot_obj, PDF_NAME(A), a);
-		}
+
+		pdf_dict_put_drop(ctx, annot_obj, PDF_NAME(A),
+			pdf_new_action_from_link(ctx, doc, uri));
 
 		/*
 			Both annotation object and annotation structure are now created.
@@ -577,10 +576,10 @@ pdf_create_link(fz_context *ctx, pdf_page *page, fz_rect bbox, const char *uri)
 	}
 	fz_always(ctx)
 	{
-		pdf_drop_obj(ctx, a);
 		pdf_drop_obj(ctx, bs);
 		pdf_drop_obj(ctx, annot_obj);
 		pdf_drop_obj(ctx, ind_obj);
+		pdf_end_operation(ctx, page->doc);
 	}
 	fz_catch(ctx)
 	{
@@ -761,40 +760,30 @@ pdf_create_annot(fz_context *ctx, pdf_page *page, enum pdf_annot_type type)
 }
 
 static int
-remove_from_tree(fz_context *ctx, pdf_obj *arr, pdf_obj *item)
+remove_from_tree(fz_context *ctx, pdf_obj *arr, pdf_obj *item, pdf_cycle_list *cycle_up)
 {
+	pdf_cycle_list cycle;
 	int i, n, res = 0;
 
-	if (arr == NULL || pdf_mark_obj(ctx, arr))
+	if (arr == NULL || pdf_cycle(ctx, &cycle, cycle_up, arr))
 		return 0;
 
-	fz_try(ctx)
+	n = pdf_array_len(ctx, arr);
+	for (i = 0; i < n; ++i)
 	{
-		n = pdf_array_len(ctx, arr);
-		for (i = 0; i < n; ++i)
+		pdf_obj *obj = pdf_array_get(ctx, arr, i);
+		if (obj == item)
 		{
-			pdf_obj *obj = pdf_array_get(ctx, arr, i);
-			if (obj == item)
-			{
-				pdf_array_delete(ctx, arr, i);
-				res = 1;
-				break;
-			}
+			pdf_array_delete(ctx, arr, i);
+			res = 1;
+			break;
+		}
 
-			if (remove_from_tree(ctx, pdf_dict_get(ctx, obj, PDF_NAME(Kids)), item))
-			{
-				res = 1;
-				break;
-			}
+		if (remove_from_tree(ctx, pdf_dict_get(ctx, obj, PDF_NAME(Kids)), item, &cycle))
+		{
+			res = 1;
+			break;
 		}
-	}
-	fz_always(ctx)
-	{
-		pdf_unmark_obj(ctx, arr);
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
 	}
 
 	return res;
@@ -874,7 +863,7 @@ pdf_delete_annot(fz_context *ctx, pdf_page *page, pdf_annot *annot)
 			pdf_obj *root = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME(Root));
 			pdf_obj *acroform = pdf_dict_get(ctx, root, PDF_NAME(AcroForm));
 			pdf_obj *fields = pdf_dict_get(ctx, acroform, PDF_NAME(Fields));
-			(void)remove_from_tree(ctx, fields, annot->obj);
+			(void)remove_from_tree(ctx, fields, annot->obj, NULL);
 		}
 
 		/* The garbage collection pass when saving will remove the annot object,
@@ -1911,7 +1900,7 @@ pdf_set_annot_vertices(fz_context *ctx, pdf_annot *annot, int n, const fz_point
 
 void pdf_clear_annot_vertices(fz_context *ctx, pdf_annot *annot)
 {
-	pdf_annot_push_local_xref(ctx, annot);
+	begin_annot_op(ctx, annot, "Clear vertices");
 
 	fz_try(ctx)
 	{
@@ -1919,7 +1908,7 @@ void pdf_clear_annot_vertices(fz_context *ctx, pdf_annot *annot)
 		pdf_dict_del(ctx, annot->obj, PDF_NAME(Vertices));
 	}
 	fz_always(ctx)
-		pdf_annot_pop_local_xref(ctx, annot);
+		end_annot_op(ctx, annot);
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 
@@ -2306,12 +2295,12 @@ pdf_set_annot_ink_list(fz_context *ctx, pdf_annot *annot, int n, const int *coun
 void
 pdf_clear_annot_ink_list(fz_context *ctx, pdf_annot *annot)
 {
-	pdf_annot_push_local_xref(ctx, annot);
+	begin_annot_op(ctx, annot, "Clear ink list");
 
 	fz_try(ctx)
 		pdf_dict_del(ctx, annot->obj, PDF_NAME(InkList));
 	fz_always(ctx)
-		pdf_annot_pop_local_xref(ctx, annot);
+		end_annot_op(ctx, annot);
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 
@@ -2817,6 +2806,8 @@ pdf_set_annot_appearance_from_display_list(fz_context *ctx, pdf_annot *annot, co
 	fz_var(contents);
 	fz_var(res);
 
+	begin_annot_op(ctx, annot, "Set appearance stream");
+
 	fz_try(ctx)
 	{
 		res = pdf_new_dict(ctx, doc, 1);
@@ -2834,7 +2825,60 @@ pdf_set_annot_appearance_from_display_list(fz_context *ctx, pdf_annot *annot, co
 		fz_drop_device(ctx, dev);
 		fz_drop_buffer(ctx, contents);
 		pdf_drop_obj(ctx, res);
+		end_annot_op(ctx, annot);
 	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 }
+
+static pdf_obj *filespec_subtypes[] = {
+	PDF_NAME(FileAttachment),
+	NULL,
+};
+
+int
+pdf_annot_has_filespec(fz_context *ctx, pdf_annot *annot)
+{
+	return is_allowed_subtype_wrap(ctx, annot, PDF_NAME(FS), filespec_subtypes);
+}
+
+pdf_obj *
+pdf_annot_filespec(fz_context *ctx, pdf_annot *annot)
+{
+	pdf_obj *filespec;
+
+	pdf_annot_push_local_xref(ctx, annot);
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(FS), filespec_subtypes);
+		filespec = pdf_dict_get(ctx, annot->obj, PDF_NAME(FS));
+	}
+	fz_always(ctx)
+		pdf_annot_pop_local_xref(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return filespec;
+}
+
+void
+pdf_set_annot_filespec(fz_context *ctx, pdf_annot *annot, pdf_obj *fs)
+{
+	if (!pdf_is_embedded_file(ctx, fs))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot set non-filespec as annotation filespec");
+
+	begin_annot_op(ctx, annot, "Set filespec");
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(M), markup_subtypes);
+		pdf_dict_put_drop(ctx, pdf_annot_obj(ctx, annot), PDF_NAME(FS), fs);
+	}
+	fz_always(ctx)
+		end_annot_op(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	pdf_dirty_annot(ctx, annot);
+}
