@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2022 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
@@ -402,6 +424,38 @@ pdf_load_substitute_cjk_font(fz_context *ctx, pdf_font_desc *fontdesc, const cha
 
 	fontdesc->font->flags.ft_substitute = 1;
 	fontdesc->font->flags.ft_stretch = 0;
+	fontdesc->font->flags.cjk = 1;
+	fontdesc->font->flags.cjk_lang = ros;
+}
+
+static struct { int ros, serif; const char *name; } known_cjk_fonts[] = {
+	{ FZ_ADOBE_GB, 0, "SimFang" },
+	{ FZ_ADOBE_GB, 0, "SimHei" },
+	{ FZ_ADOBE_GB, 1, "SimKai" },
+	{ FZ_ADOBE_GB, 1, "SimLi" },
+	{ FZ_ADOBE_GB, 1, "SimSun" },
+	{ FZ_ADOBE_GB, 1, "NSimSun" },
+	{ FZ_ADOBE_GB, 1, "Song" },
+
+	{ FZ_ADOBE_CNS, 1, "MingLiU" },
+	{ FZ_ADOBE_CNS, 1, "PMingLiU" },
+
+	{ FZ_ADOBE_JAPAN, 0, "Gothic" },
+	{ FZ_ADOBE_JAPAN, 0, "PGothic" },
+	{ FZ_ADOBE_JAPAN, 1, "Mincho" },
+	{ FZ_ADOBE_JAPAN, 1, "PMincho" },
+
+	{ FZ_ADOBE_KOREA, 1, "Batang" },
+	{ FZ_ADOBE_KOREA, 0, "Gulim" },
+	{ FZ_ADOBE_KOREA, 0, "Dotum" },
+};
+
+static int match_font_name(const char *s, const char *ref)
+{
+	/* Skip "MS-" prefix if present. */
+	if (s[0] == 'M' && s[1] == 'S' && s[2] == '-')
+		return !strncmp(s+3, ref, strlen(ref));
+	return !strncmp(s, ref, strlen(ref));
 }
 
 static void
@@ -440,8 +494,21 @@ pdf_load_system_font(fz_context *ctx, pdf_font_desc *fontdesc, const char *fontn
 			pdf_load_substitute_cjk_font(ctx, fontdesc, fontname, FZ_ADOBE_KOREA, serif);
 		else
 		{
+			size_t i;
 			if (strcmp(collection, "Adobe-Identity") != 0)
 				fz_warn(ctx, "unknown cid collection: %s", collection);
+
+			// Recognize common CJK fonts when using Identity or other non-CJK CMap
+			for (i = 0; i < nelem(known_cjk_fonts); ++i)
+			{
+				if (match_font_name(fontname, known_cjk_fonts[i].name))
+				{
+					pdf_load_substitute_cjk_font(ctx, fontdesc, fontname,
+						known_cjk_fonts[i].ros, known_cjk_fonts[i].serif);
+					return;
+				}
+			}
+
 			pdf_load_substitute_font(ctx, fontdesc, fontname, mono, serif, bold, italic);
 		}
 	}
@@ -451,14 +518,63 @@ pdf_load_system_font(fz_context *ctx, pdf_font_desc *fontdesc, const char *fontn
 	}
 }
 
+#define TTF_U16(p) ((uint16_t) ((p)[0]<<8) | ((p)[1]))
+#define TTF_U32(p) ((uint32_t) ((p)[0]<<24) | ((p)[1]<<16) | ((p)[2]<<8) | ((p)[3]))
+
+static fz_buffer *
+pdf_extract_cff_subtable(fz_context *ctx, unsigned char *data, size_t size)
+{
+	size_t num_tables = TTF_U16(data + 4);
+	size_t i;
+
+	if (12 + num_tables * 16 > size)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "invalid TTF header");
+
+	for (i = 0; i < num_tables; ++i)
+	{
+		unsigned char *record = data + 12 + i * 16;
+		if (!memcmp("CFF ", record, 4))
+		{
+			uint64_t offset = TTF_U32(record + 8);
+			uint64_t length = TTF_U32(record + 12);
+			uint64_t end = offset + length;
+			if (end > size)
+				fz_throw(ctx, FZ_ERROR_GENERIC, "invalid TTF subtable offset/length");
+			return fz_new_buffer_from_copied_data(ctx, data + offset, length);
+		}
+	}
+
+	return NULL;
+}
+
 static void
 pdf_load_embedded_font(fz_context *ctx, pdf_document *doc, pdf_font_desc *fontdesc, const char *fontname, pdf_obj *stmref)
 {
 	fz_buffer *buf;
+	unsigned char *data;
+	size_t size;
+
+	fz_var(buf);
 
 	buf = pdf_load_stream(ctx, stmref);
+
 	fz_try(ctx)
+	{
+		/* Extract CFF subtable for OpenType fonts: */
+		size = fz_buffer_storage(ctx, buf, &data);
+		if (size > 12) {
+			if (!memcmp("OTTO", data, 4)) {
+				fz_buffer *cff = pdf_extract_cff_subtable(ctx, data, size);
+				if (cff)
+				{
+					fz_drop_buffer(ctx, buf);
+					buf = cff;
+				}
+			}
+		}
+
 		fontdesc->font = fz_new_font_from_buffer(ctx, fontname, buf, 0, 1);
+	}
 	fz_always(ctx)
 		fz_drop_buffer(ctx, buf);
 	fz_catch(ctx)
@@ -607,6 +723,19 @@ select_unknown_cmap(FT_Face face)
 	return NULL;
 }
 
+static int use_s22pdf_workaround(fz_context *ctx, pdf_obj *dict, pdf_obj *descriptor)
+{
+	if (descriptor)
+	{
+		if (pdf_dict_get(ctx, dict, PDF_NAME(Encoding)) != PDF_NAME(WinAnsiEncoding))
+			return 0;
+		if (pdf_dict_get_int(ctx, descriptor, PDF_NAME(Flags)) != 4)
+			return 0;
+		return 1;
+	}
+	return 0;
+}
+
 static pdf_font_desc *
 pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 {
@@ -647,10 +776,7 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 			pdf_load_builtin_font(ctx, fontdesc, basefont, 0);
 
 		/* Some chinese documents mistakenly consider WinAnsiEncoding to be codepage 936 */
-		if (descriptor && pdf_is_string(ctx, pdf_dict_get(ctx, descriptor, PDF_NAME(FontName))) &&
-			!pdf_dict_get(ctx, dict, PDF_NAME(ToUnicode)) &&
-			pdf_name_eq(ctx, pdf_dict_get(ctx, dict, PDF_NAME(Encoding)), PDF_NAME(WinAnsiEncoding)) &&
-			pdf_dict_get_int(ctx, descriptor, PDF_NAME(Flags)) == 4)
+		if (use_s22pdf_workaround(ctx, dict, descriptor))
 		{
 			char *cp936fonts[] = {
 				"\xCB\xCE\xCC\xE5", "SimSun,Regular",
@@ -684,6 +810,9 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 		/* Encoding */
 
 		symbolic = fontdesc->flags & 4;
+		/* Bug 703273: If non-symbolic, we're not symbolic. */
+		if (fontdesc->flags & 32)
+			symbolic = 0;
 
 		if (kind == TYPE1)
 			cmap = select_type1_cmap(face);
@@ -1018,12 +1147,17 @@ load_cid_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict, pdf_obj *encodi
 			const char *reg, *ord;
 
 			cidinfo = pdf_dict_get(ctx, dict, PDF_NAME(CIDSystemInfo));
-			if (!cidinfo)
-				fz_throw(ctx, FZ_ERROR_SYNTAX, "cid font is missing info");
-
-			reg = pdf_dict_get_string(ctx, cidinfo, PDF_NAME(Registry), NULL);
-			ord = pdf_dict_get_string(ctx, cidinfo, PDF_NAME(Ordering), NULL);
-			fz_snprintf(collection, sizeof collection, "%s-%s", reg, ord);
+			if (cidinfo)
+			{
+				reg = pdf_dict_get_string(ctx, cidinfo, PDF_NAME(Registry), NULL);
+				ord = pdf_dict_get_string(ctx, cidinfo, PDF_NAME(Ordering), NULL);
+				fz_snprintf(collection, sizeof collection, "%s-%s", reg, ord);
+			}
+			else
+			{
+				fz_warn(ctx, "CIDFont is missing CIDSystemInfo dictionary; assuming Adobe-Identity");
+				fz_strlcpy(collection, "Adobe-Identity", sizeof collection);
+			}
 		}
 
 		/* Encoding */
@@ -1362,9 +1496,6 @@ pdf_load_font(fz_context *ctx, pdf_document *doc, pdf_obj *rdb, pdf_obj *dict)
 	pdf_font_desc *fontdesc = NULL;
 	int type3 = 0;
 
-	if (pdf_obj_marked(ctx, dict))
-		fz_throw(ctx, FZ_ERROR_SYNTAX, "Recursive Type3 font definition.");
-
 	if ((fontdesc = pdf_find_item(ctx, pdf_drop_font_imp, dict)) != NULL)
 	{
 		return fontdesc;
@@ -1404,7 +1535,6 @@ pdf_load_font(fz_context *ctx, pdf_document *doc, pdf_obj *rdb, pdf_obj *dict)
 		fontdesc = pdf_load_simple_font(ctx, doc, dict);
 	}
 
-	pdf_mark_obj(ctx, dict);
 	fz_try(ctx)
 	{
 		/* Create glyph width table for stretching substitute fonts and text extraction. */
@@ -1412,12 +1542,20 @@ pdf_load_font(fz_context *ctx, pdf_document *doc, pdf_obj *rdb, pdf_obj *dict)
 
 		/* Load CharProcs */
 		if (type3)
-			pdf_load_type3_glyphs(ctx, doc, fontdesc);
+		{
+			if (doc->type3_lock)
+				fz_throw(ctx, FZ_ERROR_GENERIC, "recursive type3 font");
+			doc->type3_lock = 1;
+			fz_try(ctx)
+				pdf_load_type3_glyphs(ctx, doc, fontdesc);
+			fz_always(ctx)
+				doc->type3_lock = 0;
+			fz_catch(ctx)
+				fz_rethrow(ctx);
+		}
 
 		pdf_store_item(ctx, dict, fontdesc, fontdesc->size);
 	}
-	fz_always(ctx)
-		pdf_unmark_obj(ctx, dict);
 	fz_catch(ctx)
 	{
 		pdf_drop_font(ctx, fontdesc);
