@@ -1,5 +1,27 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
-#include "mupdf/pdf.h"
+#include "pdf-annot-imp.h"
 
 #include <stdlib.h>
 #include <string.h>
@@ -21,23 +43,18 @@ int pdf_count_pages_imp(fz_context *ctx, fz_document *doc, int chapter)
 }
 
 static int
-pdf_load_page_tree_imp(fz_context *ctx, pdf_document *doc, pdf_obj *node, int idx)
+pdf_load_page_tree_imp(fz_context *ctx, pdf_document *doc, pdf_obj *node, int idx, pdf_cycle_list *cycle_up)
 {
+	pdf_cycle_list cycle;
 	pdf_obj *type = pdf_dict_get(ctx, node, PDF_NAME(Type));
 	if (pdf_name_eq(ctx, type, PDF_NAME(Pages)))
 	{
 		pdf_obj *kids = pdf_dict_get(ctx, node, PDF_NAME(Kids));
 		int i, n = pdf_array_len(ctx, kids);
-
-		if (pdf_mark_obj(ctx, node))
+		if (pdf_cycle(ctx, &cycle, cycle_up, node))
 			fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in page tree");
-		fz_try(ctx)
-			for (i = 0; i < n; ++i)
-				idx = pdf_load_page_tree_imp(ctx, doc, pdf_array_get(ctx, kids, i), idx);
-		fz_always(ctx)
-			pdf_unmark_obj(ctx, node);
-		fz_catch(ctx)
-			fz_rethrow(ctx);
+		for (i = 0; i < n; ++i)
+			idx = pdf_load_page_tree_imp(ctx, doc, pdf_array_get(ctx, kids, i), idx, &cycle);
 	}
 	else if (pdf_name_eq(ctx, type, PDF_NAME(Page)))
 	{
@@ -69,7 +86,7 @@ pdf_load_page_tree(fz_context *ctx, pdf_document *doc)
 	{
 		doc->rev_page_count = pdf_count_pages(ctx, doc);
 		doc->rev_page_map = Memento_label(fz_malloc_array(ctx, doc->rev_page_count, pdf_rev_page_map), "pdf_rev_page_map");
-		pdf_load_page_tree_imp(ctx, doc, pdf_dict_getp(ctx, pdf_trailer(ctx, doc), "Root/Pages"), 0);
+		pdf_load_page_tree_imp(ctx, doc, pdf_dict_getp(ctx, pdf_trailer(ctx, doc), "Root/Pages"), 0, NULL);
 		qsort(doc->rev_page_map, doc->rev_page_count, sizeof *doc->rev_page_map, cmp_rev_page_map);
 	}
 }
@@ -82,26 +99,15 @@ pdf_drop_page_tree(fz_context *ctx, pdf_document *doc)
 	doc->rev_page_count = 0;
 }
 
-enum
-{
-	LOCAL_STACK_SIZE = 16
-};
-
 static pdf_obj *
 pdf_lookup_page_loc_imp(fz_context *ctx, pdf_document *doc, pdf_obj *node, int *skip, pdf_obj **parentp, int *indexp)
 {
+	pdf_mark_list mark_list;
 	pdf_obj *kids;
 	pdf_obj *hit = NULL;
 	int i, len;
-	pdf_obj *local_stack[LOCAL_STACK_SIZE];
-	pdf_obj **stack = &local_stack[0];
-	int stack_max = LOCAL_STACK_SIZE;
-	int stack_len = 0;
 
-	fz_var(hit);
-	fz_var(stack);
-	fz_var(stack_len);
-	fz_var(stack_max);
+	pdf_mark_list_init(ctx, &mark_list);
 
 	fz_try(ctx)
 	{
@@ -113,23 +119,7 @@ pdf_lookup_page_loc_imp(fz_context *ctx, pdf_document *doc, pdf_obj *node, int *
 			if (len == 0)
 				fz_throw(ctx, FZ_ERROR_GENERIC, "malformed page tree");
 
-			/* Every node we need to unmark goes into the stack */
-			if (stack_len == stack_max)
-			{
-				if (stack == &local_stack[0])
-				{
-					stack = fz_malloc_array(ctx, stack_max * 2, pdf_obj*);
-					memcpy(stack, &local_stack[0], stack_max * sizeof(*stack));
-				}
-				else
-				{
-					stack = fz_realloc_array(ctx, stack, stack_max * 2, pdf_obj*);
-				}
-				stack_max *= 2;
-			}
-			stack[stack_len++] = node;
-
-			if (pdf_mark_obj(ctx, node))
+			if (pdf_mark_list_push(ctx, &mark_list, node))
 				fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in page tree");
 
 			for (i = 0; i < len; i++)
@@ -177,10 +167,7 @@ pdf_lookup_page_loc_imp(fz_context *ctx, pdf_document *doc, pdf_obj *node, int *
 	}
 	fz_always(ctx)
 	{
-		for (i = stack_len; i > 0; i--)
-			pdf_unmark_obj(ctx, stack[i-1]);
-		if (stack != &local_stack[0])
-			fz_free(ctx, stack);
+		pdf_mark_list_free(ctx, &mark_list);
 	}
 	fz_catch(ctx)
 	{
@@ -240,20 +227,24 @@ pdf_count_pages_before_kid(fz_context *ctx, pdf_document *doc, pdf_obj *parent,
 static int
 pdf_lookup_page_number_slow(fz_context *ctx, pdf_document *doc, pdf_obj *node)
 {
+	pdf_mark_list mark_list;
 	int needle = pdf_to_num(ctx, node);
 	int total = 0;
-	pdf_obj *parent, *parent2;
+	pdf_obj *parent;
 
 	if (!pdf_name_eq(ctx, pdf_dict_get(ctx, node, PDF_NAME(Type)), PDF_NAME(Page)))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "invalid page object");
+	{
+		fz_warn(ctx, "invalid page object");
+		return -1;
+	}
 
-	parent2 = parent = pdf_dict_get(ctx, node, PDF_NAME(Parent));
-	fz_var(parent);
+	pdf_mark_list_init(ctx, &mark_list);
+	parent = pdf_dict_get(ctx, node, PDF_NAME(Parent));
 	fz_try(ctx)
 	{
 		while (pdf_is_dict(ctx, parent))
 		{
-			if (pdf_mark_obj(ctx, parent))
+			if (pdf_mark_list_push(ctx, &mark_list, parent))
 				fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in page tree (parents)");
 			total += pdf_count_pages_before_kid(ctx, doc, parent, needle);
 			needle = pdf_to_num(ctx, parent);
@@ -261,20 +252,9 @@ pdf_lookup_page_number_slow(fz_context *ctx, pdf_document *doc, pdf_obj *node)
 		}
 	}
 	fz_always(ctx)
-	{
-		/* Run back and unmark */
-		while (parent2)
-		{
-			pdf_unmark_obj(ctx, parent2);
-			if (parent2 == parent)
-				break;
-			parent2 = pdf_dict_get(ctx, parent2, PDF_NAME(Parent));
-		}
-	}
+		pdf_mark_list_free(ctx, &mark_list);
 	fz_catch(ctx)
-	{
 		fz_rethrow(ctx);
-	}
 
 	return total;
 }
@@ -365,7 +345,7 @@ enum
 	PDF_FLAGS_MEMO_OP = 1
 };
 
-static int pdf_resources_use_blending(fz_context *ctx, pdf_obj *rdb);
+static int pdf_resources_use_blending(fz_context *ctx, pdf_obj *rdb, pdf_cycle_list *cycle_up);
 
 static int
 pdf_extgstate_uses_blending(fz_context *ctx, pdf_obj *dict)
@@ -377,28 +357,35 @@ pdf_extgstate_uses_blending(fz_context *ctx, pdf_obj *dict)
 }
 
 static int
-pdf_pattern_uses_blending(fz_context *ctx, pdf_obj *dict)
+pdf_pattern_uses_blending(fz_context *ctx, pdf_obj *dict, pdf_cycle_list *cycle_up)
 {
 	pdf_obj *obj;
+	pdf_cycle_list cycle;
+	if (pdf_cycle(ctx, &cycle, cycle_up, dict))
+		return 0;
 	obj = pdf_dict_get(ctx, dict, PDF_NAME(Resources));
-	if (pdf_resources_use_blending(ctx, obj))
+	if (pdf_resources_use_blending(ctx, obj, &cycle))
 		return 1;
 	obj = pdf_dict_get(ctx, dict, PDF_NAME(ExtGState));
 	return pdf_extgstate_uses_blending(ctx, obj);
 }
 
 static int
-pdf_xobject_uses_blending(fz_context *ctx, pdf_obj *dict)
+pdf_xobject_uses_blending(fz_context *ctx, pdf_obj *dict, pdf_cycle_list *cycle_up)
 {
 	pdf_obj *obj = pdf_dict_get(ctx, dict, PDF_NAME(Resources));
+	pdf_cycle_list cycle;
+	if (pdf_cycle(ctx, &cycle, cycle_up, dict))
+		return 0;
 	if (pdf_name_eq(ctx, pdf_dict_getp(ctx, dict, "Group/S"), PDF_NAME(Transparency)))
 		return 1;
-	return pdf_resources_use_blending(ctx, obj);
+	return pdf_resources_use_blending(ctx, obj, &cycle);
 }
 
 static int
-pdf_resources_use_blending(fz_context *ctx, pdf_obj *rdb)
+pdf_resources_use_blending(fz_context *ctx, pdf_obj *rdb, pdf_cycle_list *cycle_up)
 {
+	pdf_cycle_list cycle;
 	pdf_obj *obj;
 	int i, n, useBM = 0;
 
@@ -410,48 +397,37 @@ pdf_resources_use_blending(fz_context *ctx, pdf_obj *rdb)
 		return useBM;
 
 	/* stop on cyclic resource dependencies */
-	if (pdf_mark_obj(ctx, rdb))
+	if (pdf_cycle(ctx, &cycle, cycle_up, rdb))
 		return 0;
 
-	fz_try(ctx)
+	obj = pdf_dict_get(ctx, rdb, PDF_NAME(ExtGState));
+	n = pdf_dict_len(ctx, obj);
+	for (i = 0; i < n; i++)
+		if (pdf_extgstate_uses_blending(ctx, pdf_dict_get_val(ctx, obj, i)))
+			goto found;
+
+	obj = pdf_dict_get(ctx, rdb, PDF_NAME(Pattern));
+	n = pdf_dict_len(ctx, obj);
+	for (i = 0; i < n; i++)
+		if (pdf_pattern_uses_blending(ctx, pdf_dict_get_val(ctx, obj, i), &cycle))
+			goto found;
+
+	obj = pdf_dict_get(ctx, rdb, PDF_NAME(XObject));
+	n = pdf_dict_len(ctx, obj);
+	for (i = 0; i < n; i++)
+		if (pdf_xobject_uses_blending(ctx, pdf_dict_get_val(ctx, obj, i), &cycle))
+			goto found;
+	if (0)
 	{
-		obj = pdf_dict_get(ctx, rdb, PDF_NAME(ExtGState));
-		n = pdf_dict_len(ctx, obj);
-		for (i = 0; i < n; i++)
-			if (pdf_extgstate_uses_blending(ctx, pdf_dict_get_val(ctx, obj, i)))
-				goto found;
-
-		obj = pdf_dict_get(ctx, rdb, PDF_NAME(Pattern));
-		n = pdf_dict_len(ctx, obj);
-		for (i = 0; i < n; i++)
-			if (pdf_pattern_uses_blending(ctx, pdf_dict_get_val(ctx, obj, i)))
-				goto found;
-
-		obj = pdf_dict_get(ctx, rdb, PDF_NAME(XObject));
-		n = pdf_dict_len(ctx, obj);
-		for (i = 0; i < n; i++)
-			if (pdf_xobject_uses_blending(ctx, pdf_dict_get_val(ctx, obj, i)))
-				goto found;
-		if (0)
-		{
 found:
-			useBM = 1;
-		}
-	}
-	fz_always(ctx)
-	{
-		pdf_unmark_obj(ctx, rdb);
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
+		useBM = 1;
 	}
 
 	pdf_set_obj_memo(ctx, rdb, PDF_FLAGS_MEMO_BM, useBM);
 	return useBM;
 }
 
-static int pdf_resources_use_overprint(fz_context *ctx, pdf_obj *rdb);
+static int pdf_resources_use_overprint(fz_context *ctx, pdf_obj *rdb, pdf_cycle_list *cycle_up);
 
 static int
 pdf_extgstate_uses_overprint(fz_context *ctx, pdf_obj *dict)
@@ -463,26 +439,33 @@ pdf_extgstate_uses_overprint(fz_context *ctx, pdf_obj *dict)
 }
 
 static int
-pdf_pattern_uses_overprint(fz_context *ctx, pdf_obj *dict)
+pdf_pattern_uses_overprint(fz_context *ctx, pdf_obj *dict, pdf_cycle_list *cycle_up)
 {
 	pdf_obj *obj;
+	pdf_cycle_list cycle;
+	if (pdf_cycle(ctx, &cycle, cycle_up, dict))
+		return 0;
 	obj = pdf_dict_get(ctx, dict, PDF_NAME(Resources));
-	if (pdf_resources_use_overprint(ctx, obj))
+	if (pdf_resources_use_overprint(ctx, obj, &cycle))
 		return 1;
 	obj = pdf_dict_get(ctx, dict, PDF_NAME(ExtGState));
 	return pdf_extgstate_uses_overprint(ctx, obj);
 }
 
 static int
-pdf_xobject_uses_overprint(fz_context *ctx, pdf_obj *dict)
+pdf_xobject_uses_overprint(fz_context *ctx, pdf_obj *dict, pdf_cycle_list *cycle_up)
 {
 	pdf_obj *obj = pdf_dict_get(ctx, dict, PDF_NAME(Resources));
-	return pdf_resources_use_overprint(ctx, obj);
+	pdf_cycle_list cycle;
+	if (pdf_cycle(ctx, &cycle, cycle_up, dict))
+		return 0;
+	return pdf_resources_use_overprint(ctx, obj, &cycle);
 }
 
 static int
-pdf_resources_use_overprint(fz_context *ctx, pdf_obj *rdb)
+pdf_resources_use_overprint(fz_context *ctx, pdf_obj *rdb, pdf_cycle_list *cycle_up)
 {
+	pdf_cycle_list cycle;
 	pdf_obj *obj;
 	int i, n, useOP = 0;
 
@@ -494,41 +477,30 @@ pdf_resources_use_overprint(fz_context *ctx, pdf_obj *rdb)
 		return useOP;
 
 	/* stop on cyclic resource dependencies */
-	if (pdf_mark_obj(ctx, rdb))
+	if (pdf_cycle(ctx, &cycle, cycle_up, rdb))
 		return 0;
 
-	fz_try(ctx)
+	obj = pdf_dict_get(ctx, rdb, PDF_NAME(ExtGState));
+	n = pdf_dict_len(ctx, obj);
+	for (i = 0; i < n; i++)
+		if (pdf_extgstate_uses_overprint(ctx, pdf_dict_get_val(ctx, obj, i)))
+			goto found;
+
+	obj = pdf_dict_get(ctx, rdb, PDF_NAME(Pattern));
+	n = pdf_dict_len(ctx, obj);
+	for (i = 0; i < n; i++)
+		if (pdf_pattern_uses_overprint(ctx, pdf_dict_get_val(ctx, obj, i), &cycle))
+			goto found;
+
+	obj = pdf_dict_get(ctx, rdb, PDF_NAME(XObject));
+	n = pdf_dict_len(ctx, obj);
+	for (i = 0; i < n; i++)
+		if (pdf_xobject_uses_overprint(ctx, pdf_dict_get_val(ctx, obj, i), &cycle))
+			goto found;
+	if (0)
 	{
-		obj = pdf_dict_get(ctx, rdb, PDF_NAME(ExtGState));
-		n = pdf_dict_len(ctx, obj);
-		for (i = 0; i < n; i++)
-			if (pdf_extgstate_uses_overprint(ctx, pdf_dict_get_val(ctx, obj, i)))
-				goto found;
-
-		obj = pdf_dict_get(ctx, rdb, PDF_NAME(Pattern));
-		n = pdf_dict_len(ctx, obj);
-		for (i = 0; i < n; i++)
-			if (pdf_pattern_uses_overprint(ctx, pdf_dict_get_val(ctx, obj, i)))
-				goto found;
-
-		obj = pdf_dict_get(ctx, rdb, PDF_NAME(XObject));
-		n = pdf_dict_len(ctx, obj);
-		for (i = 0; i < n; i++)
-			if (pdf_xobject_uses_overprint(ctx, pdf_dict_get_val(ctx, obj, i)))
-				goto found;
-		if (0)
-		{
 found:
-			useOP = 1;
-		}
-	}
-	fz_always(ctx)
-	{
-		pdf_unmark_obj(ctx, rdb);
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
+		useOP = 1;
 	}
 
 	pdf_set_obj_memo(ctx, rdb, PDF_FLAGS_MEMO_OP, useOP);
@@ -632,7 +604,7 @@ pdf_page_obj_transform(fz_context *ctx, pdf_obj *pageobj, fz_rect *page_mediabox
 		page_mediabox = &pagebox;
 
 	obj = pdf_dict_get(ctx, pageobj, PDF_NAME(UserUnit));
-	if (pdf_is_real(ctx, obj))
+	if (pdf_is_number(ctx, obj))
 		userunit = pdf_to_real(ctx, obj);
 
 	mediabox = pdf_to_rect(ctx, pdf_dict_get_inheritable(ctx, pageobj, PDF_NAME(MediaBox)));
@@ -688,11 +660,18 @@ pdf_page_transform(fz_context *ctx, pdf_page *page, fz_rect *page_mediabox, fz_m
 }
 
 static void
-find_seps(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_obj *clearme)
+find_seps(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_mark_list *clearme)
 {
 	int i, n;
 	pdf_obj *nameobj, *cols;
 
+	if (!obj)
+		return;
+
+	// Already seen this ColorSpace...
+	if (pdf_mark_list_push(ctx, clearme, obj))
+		return;
+
 	nameobj = pdf_array_get(ctx, obj, 0);
 	if (pdf_name_eq(ctx, nameobj, PDF_NAME(Separation)))
 	{
@@ -735,26 +714,10 @@ find_seps(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_obj *clearme
 	}
 	else if (pdf_name_eq(ctx, nameobj, PDF_NAME(Indexed)))
 	{
-		if (pdf_is_indirect(ctx, obj))
-		{
-			if (pdf_mark_obj(ctx, obj))
-				return; /* already been here */
-			/* remember to clear this colorspace dictionary at the end */
-			pdf_array_push(ctx, clearme, obj);
-		}
-
 		find_seps(ctx, seps, pdf_array_get(ctx, obj, 1), clearme);
 	}
 	else if (pdf_name_eq(ctx, nameobj, PDF_NAME(DeviceN)))
 	{
-		if (pdf_is_indirect(ctx, obj))
-		{
-			if (pdf_mark_obj(ctx, obj))
-				return; /* already been here */
-			/* remember to clear this colorspace dictionary at the end */
-			pdf_array_push(ctx, clearme, obj);
-		}
-
 		/* If the separation colorants exists for this DeviceN color space
 		 * add those prior to our search for DeviceN color */
 		cols = pdf_dict_get(ctx, pdf_array_get(ctx, obj, 4), PDF_NAME(Colorants));
@@ -765,12 +728,19 @@ find_seps(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_obj *clearme
 }
 
 static void
-find_devn(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_obj *clearme)
+find_devn(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_mark_list *clearme)
 {
 	int i, j, n, m;
 	pdf_obj *arr;
 	pdf_obj *nameobj = pdf_array_get(ctx, obj, 0);
 
+	if (!obj)
+		return;
+
+	// Already seen this ColorSpace...
+	if (pdf_mark_list_push(ctx, clearme, obj))
+		return;
+
 	if (!pdf_name_eq(ctx, nameobj, PDF_NAME(DeviceN)))
 		return;
 
@@ -820,20 +790,21 @@ find_devn(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_obj *clearme
 	}
 }
 
-typedef void (res_finder_fn)(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_obj *clearme);
+typedef void (res_finder_fn)(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_mark_list *clearme);
 
 static void
-scan_page_seps(fz_context *ctx, pdf_obj *res, fz_separations **seps, res_finder_fn *fn, pdf_obj *clearme)
+scan_page_seps(fz_context *ctx, pdf_obj *res, fz_separations **seps, res_finder_fn *fn, pdf_mark_list *clearme)
 {
 	pdf_obj *dict;
 	pdf_obj *obj;
 	int i, n;
 
-	if (pdf_mark_obj(ctx, res))
-		return; /* already been here */
+	if (!res)
+		return;
 
-	/* remember to clear this resource dictionary at the end */
-	pdf_array_push(ctx, clearme, res);
+	// Already seen this Resources...
+	if (pdf_mark_list_push(ctx, clearme, res))
+		return;
 
 	dict = pdf_dict_get(ctx, res, PDF_NAME(ColorSpace));
 	n = pdf_dict_len(ctx, dict);
@@ -856,9 +827,13 @@ scan_page_seps(fz_context *ctx, pdf_obj *res, fz_separations **seps, res_finder_
 	for (i = 0; i < n; i++)
 	{
 		obj = pdf_dict_get_val(ctx, dict, i);
-		fn(ctx, seps, pdf_dict_get(ctx, obj, PDF_NAME(ColorSpace)), clearme);
-		/* Recurse on XObject forms. */
-		scan_page_seps(ctx, pdf_dict_get(ctx, obj, PDF_NAME(Resources)), seps, fn, clearme);
+		// Already seen this XObject...
+		if (!pdf_mark_list_push(ctx, clearme, obj))
+		{
+			fn(ctx, seps, pdf_dict_get(ctx, obj, PDF_NAME(ColorSpace)), clearme);
+			/* Recurse on XObject forms. */
+			scan_page_seps(ctx, pdf_dict_get(ctx, obj, PDF_NAME(Resources)), seps, fn, clearme);
+		}
 	}
 }
 
@@ -866,46 +841,36 @@ fz_separations *
 pdf_page_separations(fz_context *ctx, pdf_page *page)
 {
 	pdf_obj *res = pdf_page_resources(ctx, page);
-	pdf_obj *clearme = NULL;
+	pdf_mark_list clearme;
 	fz_separations *seps = NULL;
 
-	clearme = pdf_new_array(ctx, page->doc, 100);
+	pdf_mark_list_init(ctx, &clearme);
 	fz_try(ctx)
 	{
 		/* Run through and look for separations first. This is
 		 * because separations are simplest to deal with, and
 		 * because DeviceN may be implemented on top of separations.
 		 */
-		scan_page_seps(ctx, res, &seps, find_seps, clearme);
+		scan_page_seps(ctx, res, &seps, find_seps, &clearme);
 	}
 	fz_always(ctx)
-	{
-		int i, n = pdf_array_len(ctx, clearme);
-		for (i = 0; i < n; ++i)
-			pdf_unmark_obj(ctx, pdf_array_get(ctx, clearme, i));
-		pdf_drop_obj(ctx, clearme);
-	}
+		pdf_mark_list_free(ctx, &clearme);
 	fz_catch(ctx)
 	{
 		fz_drop_separations(ctx, seps);
 		fz_rethrow(ctx);
 	}
 
-	clearme = pdf_new_array(ctx, page->doc, 100);
+	pdf_mark_list_init(ctx, &clearme);
 	fz_try(ctx)
 	{
 		/* Now run through again, and look for DeviceNs. These may
 		 * have spot colors in that aren't defined in terms of
 		 * separations. */
-		scan_page_seps(ctx, res, &seps, find_devn, clearme);
+		scan_page_seps(ctx, res, &seps, find_devn, &clearme);
 	}
 	fz_always(ctx)
-	{
-		int i, n = pdf_array_len(ctx, clearme);
-		for (i = 0; i < n; ++i)
-			pdf_unmark_obj(ctx, pdf_array_get(ctx, clearme, i));
-		pdf_drop_obj(ctx, clearme);
-	}
+		pdf_mark_list_free(ctx, &clearme);
 	fz_catch(ctx)
 	{
 		fz_drop_separations(ctx, seps);
@@ -927,18 +892,15 @@ pdf_drop_page_imp(fz_context *ctx, pdf_page *page)
 	fz_drop_link(ctx, page->links);
 	pdf_drop_annots(ctx, page->annots);
 	pdf_drop_widgets(ctx, page->widgets);
-
 	pdf_drop_obj(ctx, page->obj);
-
-	fz_drop_document(ctx, &page->doc->super);
 }
 
 static pdf_page *
 pdf_new_page(fz_context *ctx, pdf_document *doc)
 {
-	pdf_page *page = fz_new_derived_page(ctx, pdf_page);
+	pdf_page *page = fz_new_derived_page(ctx, pdf_page, (fz_document*) doc);
 
-	page->doc = (pdf_document*) fz_keep_document(ctx, &doc->super);
+	page->doc = doc; /* typecast alias for page->super.doc */
 
 	page->super.drop_page = (fz_page_drop_page_fn*)pdf_drop_page_imp;
 	page->super.load_links = (fz_page_load_links_fn*)pdf_load_links;
@@ -949,6 +911,7 @@ pdf_new_page(fz_context *ctx, pdf_document *doc)
 	page->super.page_presentation = (fz_page_page_presentation_fn*)pdf_page_presentation;
 	page->super.separations = (fz_page_separations_fn *)pdf_page_separations;
 	page->super.overprint = (fz_page_uses_overprint_fn *)pdf_page_uses_overprint;
+	page->super.create_link = (fz_page_create_link_fn *)pdf_create_link;
 
 	page->obj = NULL;
 
@@ -1068,6 +1031,13 @@ pdf_update_default_colorspaces(fz_context *ctx, fz_default_colorspaces *old_cs,
 pdf_page *
 pdf_load_page(fz_context *ctx, pdf_document *doc, int number)
 {
+	return (pdf_page*)fz_load_page(ctx, (fz_document*)doc, number);
+}
+
+fz_page *
+pdf_load_page_imp(fz_context *ctx, fz_document *doc_, int chapter, int number)
+{
+	pdf_document *doc = (pdf_document*)doc_;
 	pdf_page *page;
 	pdf_annot *annot;
 	pdf_obj *pageobj, *obj;
@@ -1115,16 +1085,52 @@ pdf_load_page(fz_context *ctx, pdf_document *doc, int number)
 		pdf_obj *resources = pdf_page_resources(ctx, page);
 		if (pdf_name_eq(ctx, pdf_dict_getp(ctx, pageobj, "Group/S"), PDF_NAME(Transparency)))
 			page->transparency = 1;
-		else if (pdf_resources_use_blending(ctx, resources))
+		else if (pdf_resources_use_blending(ctx, resources, NULL))
 			page->transparency = 1;
-		for (annot = page->annots; annot && !page->transparency; annot = annot->next)
-			if (annot->ap && pdf_resources_use_blending(ctx, pdf_xobject_resources(ctx, annot->ap)))
-				page->transparency = 1;
-		if (pdf_resources_use_overprint(ctx, resources))
+		if (pdf_resources_use_overprint(ctx, resources, NULL))
 			page->overprint = 1;
-		for (annot = page->annots; annot && !page->overprint; annot = annot->next)
-			if (annot->ap && pdf_resources_use_overprint(ctx, pdf_xobject_resources(ctx, annot->ap)))
-				page->overprint = 1;
+		for (annot = page->annots; annot && !page->transparency; annot = annot->next)
+		{
+			fz_try(ctx)
+			{
+				pdf_obj *ap;
+				pdf_obj *res;
+				pdf_annot_push_local_xref(ctx, annot);
+				ap = pdf_annot_ap(ctx, annot);
+				if (!ap)
+					break;
+				res = pdf_xobject_resources(ctx, ap);
+				if (pdf_resources_use_blending(ctx, res, NULL))
+					page->transparency = 1;
+				if (pdf_resources_use_overprint(ctx, pdf_xobject_resources(ctx, res), NULL))
+					page->overprint = 1;
+			}
+			fz_always(ctx)
+				pdf_annot_pop_local_xref(ctx, annot);
+			fz_catch(ctx)
+				fz_rethrow(ctx);
+		}
+		for (annot = page->widgets; annot && !page->transparency; annot = annot->next)
+		{
+			fz_try(ctx)
+			{
+				pdf_obj *ap;
+				pdf_obj *res;
+				pdf_annot_push_local_xref(ctx, annot);
+				ap = pdf_annot_ap(ctx, annot);
+				if (!ap)
+					break;
+				res = pdf_xobject_resources(ctx, ap);
+				if (pdf_resources_use_blending(ctx, res, NULL))
+					page->transparency = 1;
+				if (pdf_resources_use_overprint(ctx, pdf_xobject_resources(ctx, res), NULL))
+					page->overprint = 1;
+			}
+			fz_always(ctx)
+				pdf_annot_pop_local_xref(ctx, annot);
+			fz_catch(ctx)
+				fz_rethrow(ctx);
+		}
 	}
 	fz_catch(ctx)
 	{
@@ -1136,12 +1142,7 @@ pdf_load_page(fz_context *ctx, pdf_document *doc, int number)
 		page->super.incomplete = 1;
 	}
 
-	return page;
-}
-
-fz_page *pdf_load_page_imp(fz_context *ctx, fz_document *doc, int chapter, int number)
-{
-	return (fz_page*)pdf_load_page(ctx, (pdf_document*)doc, number);
+	return (fz_page*)page;
 }
 
 void
