@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 /*
  * muraster -- Convert a document to a raster file.
  *
@@ -481,7 +503,7 @@ static struct {
 
 #define stringify(A) #A
 
-static void usage(void)
+static int usage(void)
 {
 	fprintf(stderr,
 		"muraster version " FZ_VERSION "\n"
@@ -520,7 +542,7 @@ static void usage(void)
 		"\n"
 		"\tpages\tcomma separated list of page numbers and ranges\n"
 		);
-	exit(1);
+	return 1;
 }
 
 static int gettime(void)
@@ -543,6 +565,8 @@ static int drawband(fz_context *ctx, fz_page *page, fz_display_list *list, fz_ma
 
 	*bit = NULL;
 
+	fz_var(dev);
+
 	fz_try(ctx)
 	{
 		fz_clear_pixmap_with_value(ctx, pix, 255);
@@ -613,11 +637,11 @@ static int dodrawpage(fz_context *ctx, int pagenum, fz_cookie *cookie, render_de
 					ibounds.y1 = ibounds.y0 + remaining_height;
 				remaining_height -= band_height;
 				w->pix = fz_new_pixmap_with_bbox(ctx, colorspace, ibounds, NULL, 0);
+				w->pix->y += band * band_height;
 				fz_set_pixmap_resolution(ctx, w->pix, x_resolution, y_resolution);
 				DEBUG_THREADS(("Worker %d, Pre-triggering band %d\n", band, band));
 				w->started = 1;
 				mu_trigger_semaphore(&w->start);
-				ctm.f -= band_height;
 			}
 			pix = workers[0].pix;
 		}
@@ -677,7 +701,8 @@ static int dodrawpage(fz_context *ctx, int pagenum, fz_cookie *cookie, render_de
 				w->started = 1;
 				mu_trigger_semaphore(&w->start);
 			}
-			ctm.f -= draw_height;
+			if (render->num_workers <= 0)
+				pix += draw_height;
 		}
 	}
 	fz_always(ctx)
@@ -780,6 +805,8 @@ static int try_render_page(fz_context *ctx, int pagenum, fz_cookie *cookie, int
 		break;
 	}
 
+	fz_close_band_writer(ctx, render->bander);
+
 	fz_drop_page(ctx, render->page);
 	fz_drop_display_list(ctx, render->list);
 	fz_drop_band_writer(ctx, render->bander);
@@ -978,8 +1005,9 @@ initialise_banding(fz_context *ctx, render_details *render, int color)
 
 	w = render->ibounds.x1 - render->ibounds.x0;
 	min_band_mem = (size_t)bpp * w * min_band_height;
-	reps = (int)(max_band_memory / min_band_mem);
-	if (reps < 1)
+	if (min_band_mem > 0)
+		reps = (int)(max_band_memory / min_band_mem);
+	if (min_band_mem == 0 || reps < 1)
 		reps = 1;
 
 	/* Adjust reps to even out the work between threads */
@@ -1442,7 +1470,7 @@ int main(int argc, char **argv)
 	{
 		switch (c)
 		{
-		default: usage(); break;
+		default: return usage();
 
 		case 'p': password = fz_optarg; break;
 
@@ -1506,7 +1534,7 @@ int main(int argc, char **argv)
 		height = y_resolution * PAPER_HEIGHT;
 
 	if (fz_optind == argc)
-		usage();
+		return usage();
 
 	if (min_band_height <= 0)
 	{
